import time

from ophyd import Signal, SignalRO, Device
import pyvisa
import re

rm = pyvisa.ResourceManager()
open_resources = {}


def list_resources():
    """Gives the results of `ResourceManager.list_resources`."""
    return rm.list_resources()


def close_resources():
    """Goes through all the opened resources and closes them."""
    for res in open_resources:
        open_resources[res].close()
    open_resources.clear()


class VISA_Signal(Signal):
    """ """

    def __init__(
        self,
        name,
        value=0.0,
        timestamp=None,
        parent=None,
        labels=None,
        kind="hinted",
        tolerance=None,
        rtolerance=None,
        metadata=None,
        cl=None,
        attr_name="",
        write=None,
        parse=None,
        parse_return_type=None,
        retry_on_error=0,
        write_delay=0,
        retry_on_timeout=False,
    ):
        """
        Parameters
        ----------
        write : str or function with str as a return value
            (Default = None)
            Determines what is written to the instrument. As a string it should have '{value}' in it (could include further formatting). The given value will be put into the string there, e.g. if `write='VOLT{value:03d}'`, the value will be given as an integer of three digits with leading zeroes.
            If `write` is a function, it should return a string. Every time a value is set, the function will be executed and the resulting string will be sent to the instrument.
            If it is None, the value will be simply converted to a string.

        parse : str or function
            (Default value = None)
            This gives how the returned string from the instrument is being parsed.
            If str, it should be a regular expression. The first result from the regex will become the new value.
            If function, it will take the returned string and the new value will be the value returned from this function.
            If `parse` is not None, a query to the instrument will be executed, instead of a simple write. If it is used to just clear the buffer, it could be an empty string.

        parse_return_type : str, class or None
            (Default = None)
            If `parse` is a string or None, the returned string (either from the instrument or from `parse`) will be converted to the given class.
            Supported strings for class names are: 'str', 'float', 'int', 'bool'.
            If `parse_return_type` and `parse` is not None, the value will become the readback for this Signal. Otherwise the value given with put becomes the readback.

        retry_on_error : int
            (Default = 0)
            The number of times the reading / setting should be tried again if
            there is an error (e.g. a VISA-IO-error).
        """
        super().__init__(
            name=name,
            value=value,
            timestamp=timestamp,
            parent=parent,
            labels=labels,
            kind=kind,
            tolerance=tolerance,
            rtolerance=rtolerance,
            metadata=metadata,
            cl=cl,
            attr_name=attr_name,
        )
        self.visa_instrument = None
        self.write = write
        self.write_delay = write_delay
        self.parse = parse
        self.retry_on_error = retry_on_error
        self.retry_on_timeout = retry_on_timeout
        if parse_return_type == "str":
            self.parse_return_type = str
        elif parse_return_type == "float":
            self.parse_return_type = float
        elif parse_return_type == "int":
            self.parse_return_type = int
        elif parse_return_type == "bool":
            self.parse_return_type = bool
        else:
            self.parse_return_type = parse_return_type

    def change_instrument(self, resource_name):
        """
        Changes the visa_instrument to the one with the given `resource_name`.

        Parameters
        ----------
        resource_name : str
            Name of the new instrument to be used.
        """
        if resource_name:
            if resource_name in open_resources:
                self.visa_instrument = open_resources[resource_name]
            else:
                self.visa_instrument = rm.open_resource(resource_name)
                open_resources[resource_name] = self.visa_instrument

    def put(self, value, *, timestamp=None, force=False, metadata=None, **kwargs):
        """
        Overwrites `ophyd.Signal.put`. The value is converted to a string, using
        `self.write`. If `self.parse` is not None, a query instead of a simple
        write is being performed, in that case, the value given to super().put
        is the returned value from the query.
        """
        if not self.write:
            write_text = str(value)
        elif isinstance(self.write, str):
            val = value
            if "value:g" in self.write:
                val = float(value)
            elif re.search(r"{value:\.\df}", self.write) is not None:
                val = float(value)
            elif "value:d" in self.write:
                val = int(value)
            write_text = self.write.format(value=val)
        else:
            write_text = self.write(value)
        if self.parse is not None:
            val = retry_query_or_write(
                write_text,
                self.visa_instrument,
                self.retry_on_error,
                write_delay=self.write_delay,
                retry_on_timeout=self.retry_on_timeout,
            )
            if self.parse:
                try:
                    if isinstance(self.parse, str):
                        val = re.match(self.parse, val).group(1)
                    else:
                        val = self.parse(val)
                except Exception as e:
                    print(e)
            if self.parse_return_type:
                try:
                    value = self.parse_return_type(val)
                except:
                    value = val
        else:
            retry_query_or_write(
                write_text,
                self.visa_instrument,
                self.retry_on_error,
                True,
                write_delay=self.write_delay,
                retry_on_timeout=self.retry_on_timeout,
            )
        super().put(
            value, timestamp=timestamp, force=force, metadata=metadata, **kwargs
        )

    def describe(self):
        """Adding "VISA" as the source for the description."""
        info = super().describe()
        info[self.name]["source"] = "VISA"
        # info[self.name].update(self.metadata)
        return info


def retry_query_or_write(
    write_text,
    visa_instrument,
    retries,
    just_write=False,
    write_delay=0,
    retry_on_timeout=False,
):
    excs = []
    while visa_instrument.currently_reading:
        time.sleep(0.001)
    visa_instrument.currently_reading = True
    if write_delay:
        time.sleep(write_delay)
    for i in range(retries + 1):
        try:
            if retry_on_timeout:
                try:
                    return single_query_or_write(
                        write_text, visa_instrument, just_write
                    )
                except pyvisa.errors.VisaIOError as e:
                    visa_instrument.clear()
                    return single_query_or_write(
                        write_text, visa_instrument, just_write
                    )
            else:
                return single_query_or_write(write_text, visa_instrument, just_write)
        except Exception as e:
            if i == retries:
                print(excs)
                raise e
            excs.append(e)


def single_query_or_write(write_text, visa_instrument, just_write=False):
    val = None
    if just_write:
        visa_instrument.write(write_text)
    else:
        val = visa_instrument.query(write_text)
    visa_instrument.currently_reading = False
    return val


class VISA_Signal_RO(SignalRO):
    """ """

    def __init__(
        self,
        name,
        value=0.0,
        timestamp=None,
        parent=None,
        labels=None,
        kind="hinted",
        tolerance=None,
        rtolerance=None,
        metadata=None,
        cl=None,
        attr_name="",
        query="",
        parse=None,
        parse_return_type="float",
        retry_on_error=0,
        write_delay=0,
        retry_on_timeout=False,
    ):
        """
        Parameters
        ----------
        query : str or function with str as a return value
            Determines what is written to the instrument. If it is a string, that string will be sent to the instrument. If it is a function, its return value will be sent to the instrument.

        parse : str or function
            (Default value = None)
            This gives how the returned string from the instrument is being parsed.
            If str, it should be a regular expression. The first result from the regex will become the new value.
            If function, it will take the returned string and the new value will be the value returned from this function.
            If `parse` is not None, a query to the instrument will be executed, instead of a simple write. If it is used to just clear the buffer, it could be an empty string.

        parse_return_type : str, class
            (Default = 'float')
            If `parse` is a string or None, the returned string (either from the instrument or from `parse`) will be converted to the given class.
            Supported strings for class names are: 'str', 'float', 'int', 'bool'.

        retry_on_error : int
            (Default = 0)
            The number of times the reading / setting should be tried again if
            there is an error (e.g. a VISA-IO-error).
        """
        super().__init__(
            name=name,
            value=value,
            timestamp=timestamp,
            parent=parent,
            labels=labels,
            kind=kind,
            tolerance=tolerance,
            rtolerance=rtolerance,
            metadata=metadata,
            cl=cl,
            attr_name=attr_name,
        )
        self.visa_instrument = None
        self.query = query
        self.parse = parse
        self.write_delay = write_delay
        self.retry_on_error = retry_on_error
        self.retry_on_timeout = retry_on_timeout
        if parse_return_type == "str":
            self.parse_return_type = str
        elif parse_return_type == "float":
            self.parse_return_type = float
        elif parse_return_type == "int":
            self.parse_return_type = int
        elif parse_return_type == "bool":
            self.parse_return_type = bool
        else:
            self.parse_return_type = parse_return_type

    def get(self):
        """
        Overwrites `ophyd.SignalRO.get`, performing a query to the instrument.
        The returned string is parsed regarding `self.parse` and converted to
        the datatype specified by `self.parse_return_type`.
        """
        old_value = self._readback
        if isinstance(self.query, str):
            query = self.query
        else:
            query = self.query()
        val = retry_query_or_write(
            query,
            self.visa_instrument,
            self.retry_on_error,
            write_delay=self.write_delay,
            retry_on_timeout=self.retry_on_timeout,
        )
        if self.parse is not None:
            try:
                if isinstance(self.parse, str):
                    val = re.match(self.parse, val).group(1)
                else:
                    val = self.parse(val)
            except Exception as e:
                print(e)
        if self.parse_return_type:
            try:
                val = self.parse_return_type(val)
            except:
                val = val
        self._readback = val
        self._run_subs(
            sub_type=self.SUB_VALUE,
            old_value=old_value,
            value=val,
            timestamp=time.time(),
        )
        return super().get()

    def describe(self):
        """Adding "VISA" as the source for the description."""
        info = super().describe()
        info[self.name]["source"] = "VISA"
        # info[self.name].update(self.metadata)
        return info


class VISA_Device(Device):
    """
    Subclasses ophyd's `Device` class. Automatically opens the specified VISA
    resource.

    Parameters
    ----------
    resource_name : str
        The name of the VISA-resource.
    read_termination : str
        (Default = '\\r\\n')
        The line termination for reading from the instrument.
    write_termination : str
        (Default = '\\r\\n')
        The line termination for writing to the instrument.
    baud_rate : int
        (Default = 9600)
        The communication baud rate.
    retry_on_error : int
        (Default = 0)
        The number of times the reading / setting should be tried again if
        there is an error (e.g. a VISA-IO-error). It is handed to all components
        that do not have a retry_on_error already defined.
    """

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        resource_name="",
        read_termination="\r\n",
        write_termination="\r\n",
        baud_rate=9600,
        timeout=2000,
        retry_on_error=0,
        retry_on_timeout=False,
        **kwargs,
    ):
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self.visa_instrument = None
        if resource_name:
            if resource_name in open_resources:
                self.visa_instrument = open_resources[resource_name]
            else:
                self.visa_instrument = rm.open_resource(resource_name)
                open_resources[resource_name] = self.visa_instrument
            self.visa_instrument.write_termination = write_termination
            self.visa_instrument.read_termination = read_termination
            self.visa_instrument.baud_rate = baud_rate
            self.visa_instrument.currently_reading = False
            self.visa_instrument.timeout = timeout

        for comp in self.walk_signals():
            it = comp.item
            it.visa_instrument = self.visa_instrument
            if hasattr(it, "retry_on_error") and not it.retry_on_error:
                it.retry_on_error = retry_on_error
                it.retry_on_timeout = retry_on_timeout



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'protocol_view.ui'
##
## Created by: Qt User Interface Compiler version 6.5.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QFrame, QGridLayout, QHeaderView,
    QLabel, QPushButton, QSizePolicy, QSplitter,
    QToolButton, QTreeView, QWidget)

class Ui_Protocol_View(object):
    def setupUi(self, Protocol_View):
        if not Protocol_View.objectName():
            Protocol_View.setObjectName(u"Protocol_View")
        Protocol_View.resize(617, 522)
        self.gridLayout = QGridLayout(Protocol_View)
        self.gridLayout.setObjectName(u"gridLayout")
        self.meas_splitter = QSplitter(Protocol_View)
        self.meas_splitter.setObjectName(u"meas_splitter")
        self.meas_splitter.setFrameShape(QFrame.NoFrame)
        self.meas_splitter.setLineWidth(10)
        self.meas_splitter.setMidLineWidth(10)
        self.meas_splitter.setOrientation(Qt.Horizontal)
        self.sequence_main_widget = QWidget(self.meas_splitter)
        self.sequence_main_widget.setObjectName(u"sequence_main_widget")
        self.gridLayout_8 = QGridLayout(self.sequence_main_widget)
        self.gridLayout_8.setObjectName(u"gridLayout_8")
        self.gridLayout_8.setContentsMargins(-1, 0, -1, 0)
        self.pushButton_remove_step = QPushButton(self.sequence_main_widget)
        self.pushButton_remove_step.setObjectName(u"pushButton_remove_step")
        self.pushButton_remove_step.setMinimumSize(QSize(30, 23))
        self.pushButton_remove_step.setMaximumSize(QSize(30, 30))
        self.pushButton_remove_step.setStyleSheet(u"QPushButton {\n"
"                                background-color:  #FF3333; \n"
"                                color: white; \n"
"                                border: none; \n"
"                                padding: 0px; \n"
"                                padding-bottom: 5px;\n"
"                                text-align: center; \n"
"                                text-decoration: none; \n"
"                                font-size: 18px; \n"
"                                margin: 2px 2px; \n"
"                                border-radius: 6px;\n"
"								font-weight: bold;\n"
"                            }\n"
"\n"
"            QPushButton:hover {\n"
"                background-color: #B22222;\n"
"            }\n"
"QPushButton:disabled {\n"
"        background-color: #808080;\n"
"    }")

        self.gridLayout_8.addWidget(self.pushButton_remove_step, 3, 2, 1, 1)

        self.toolButton_add_step = QToolButton(self.sequence_main_widget)
        self.toolButton_add_step.setObjectName(u"toolButton_add_step")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.toolButton_add_step.sizePolicy().hasHeightForWidth())
        self.toolButton_add_step.setSizePolicy(sizePolicy)
        self.toolButton_add_step.setMinimumSize(QSize(30, 23))
        self.toolButton_add_step.setMaximumSize(QSize(30, 30))
        self.toolButton_add_step.setStyleSheet(u"QToolButton {\n"
"                                background-color: #4CAF50; \n"
"                                color: white; \n"
"                                border: none; \n"
"                                padding: 0px; \n"
"                                padding-bottom: 5px;\n"
"                                text-align: center; \n"
"                                text-decoration: none; \n"
"                                font-size: 18px; \n"
"                                margin: 2px 2px; \n"
"                                border-radius: 6px;\n"
"								font-weight: bold;\n"
"                            }\n"
"\n"
"                            QToolButton:hover {\n"
"                                background-color: #45a049;\n"
"                            }")

        self.gridLayout_8.addWidget(self.toolButton_add_step, 2, 2, 1, 1)

        self.label_sequence = QLabel(self.sequence_main_widget)
        self.label_sequence.setObjectName(u"label_sequence")
        self.label_sequence.setMaximumSize(QSize(16777215, 19))
        font = QFont()
        font.setPointSize(12)
        font.setBold(True)
        self.label_sequence.setFont(font)

        self.gridLayout_8.addWidget(self.label_sequence, 0, 0, 1, 1)

        self.pushButton_move_step_up = QPushButton(self.sequence_main_widget)
        self.pushButton_move_step_up.setObjectName(u"pushButton_move_step_up")
        self.pushButton_move_step_up.setStyleSheet(u"QPushButton {\n"
"    background-color: #2a4cdf; /* This is a lighter shade of blue */\n"
"    color: white;\n"
"    border: none;\n"
"    padding: 2px 10px;\n"
"    text-align: center;\n"
"    text-decoration: none;\n"
"    font-size: 13px;\n"
"    margin: 2px 2px;\n"
"    border-radius: 6px;\n"
"    font-weight: bold;\n"
"}\n"
"\n"
"QPushButton:hover {\n"
"    background-color: #1a3cbf; /* This is a darker shade of the lighter blue for hover effect */\n"
"}\n"
"QPushButton:disabled {\n"
"        background-color: #808080;\n"
"    }")

        self.gridLayout_8.addWidget(self.pushButton_move_step_up, 2, 0, 1, 1)

        self.treeView_protocol_sequence = QTreeView(self.sequence_main_widget)
        self.treeView_protocol_sequence.setObjectName(u"treeView_protocol_sequence")

        self.gridLayout_8.addWidget(self.treeView_protocol_sequence, 5, 0, 1, 3)

        self.pushButton_move_step_down = QPushButton(self.sequence_main_widget)
        self.pushButton_move_step_down.setObjectName(u"pushButton_move_step_down")
        self.pushButton_move_step_down.setStyleSheet(u"QPushButton {\n"
"    background-color: #2a4cdf; /* This is a lighter shade of blue */\n"
"    color: white;\n"
"    border: none;\n"
"    padding: 2px 10px;\n"
"    text-align: center;\n"
"    text-decoration: none;\n"
"    font-size: 13px;\n"
"    margin: 2px 2px;\n"
"    border-radius: 6px;\n"
"    font-weight: bold;\n"
"}\n"
"\n"
"QPushButton:hover {\n"
"    background-color: #1a3cbf; /* This is a darker shade of the lighter blue for hover effect */\n"
"}\n"
"QPushButton:disabled {\n"
"        background-color: #808080;\n"
"    }")

        self.gridLayout_8.addWidget(self.pushButton_move_step_down, 3, 0, 1, 1)

        self.pushButton_move_step_out = QPushButton(self.sequence_main_widget)
        self.pushButton_move_step_out.setObjectName(u"pushButton_move_step_out")
        self.pushButton_move_step_out.setStyleSheet(u"QPushButton {\n"
"    background-color: #2a4cdf; /* This is a lighter shade of blue */\n"
"    color: white;\n"
"    border: none;\n"
"    padding: 2px 10px;\n"
"    text-align: center;\n"
"    text-decoration: none;\n"
"    font-size: 13px;\n"
"    margin: 2px 2px;\n"
"    border-radius: 6px;\n"
"    font-weight: bold;\n"
"}\n"
"\n"
"QPushButton:hover {\n"
"    background-color: #1a3cbf; /* This is a darker shade of the lighter blue for hover effect */\n"
"}\n"
"QPushButton:disabled {\n"
"        background-color: #808080;\n"
"    }")

        self.gridLayout_8.addWidget(self.pushButton_move_step_out, 3, 1, 1, 1)

        self.pushButton_move_step_in = QPushButton(self.sequence_main_widget)
        self.pushButton_move_step_in.setObjectName(u"pushButton_move_step_in")
        self.pushButton_move_step_in.setStyleSheet(u"QPushButton {\n"
"    background-color: #2a4cdf; /* This is a lighter shade of blue */\n"
"    color: white;\n"
"    border: none;\n"
"    padding: 2px 10px;\n"
"    text-align: center;\n"
"    text-decoration: none;\n"
"    font-size: 13px;\n"
"    margin: 2px 2px;\n"
"    border-radius: 6px;\n"
"    font-weight: bold;\n"
"}\n"
"\n"
"QPushButton:hover {\n"
"    background-color: #1a3cbf; /* This is a darker shade of the lighter blue for hover effect */\n"
"}\n"
"QPushButton:disabled {\n"
"        background-color: #808080;\n"
"    }")

        self.gridLayout_8.addWidget(self.pushButton_move_step_in, 2, 1, 1, 1)

        self.meas_splitter.addWidget(self.sequence_main_widget)
        self.configuration_main_widget = QWidget(self.meas_splitter)
        self.configuration_main_widget.setObjectName(u"configuration_main_widget")
        self.gridLayout_9 = QGridLayout(self.configuration_main_widget)
        self.gridLayout_9.setObjectName(u"gridLayout_9")
        self.gridLayout_9.setContentsMargins(-1, 0, 0, 0)
        self.label_configuration = QLabel(self.configuration_main_widget)
        self.label_configuration.setObjectName(u"label_configuration")
        self.label_configuration.setMaximumSize(QSize(16777215, 19))
        self.label_configuration.setFont(font)

        self.gridLayout_9.addWidget(self.label_configuration, 0, 0, 1, 1)

        self.loopstep_configuration_widget = QWidget(self.configuration_main_widget)
        self.loopstep_configuration_widget.setObjectName(u"loopstep_configuration_widget")
        self.gridLayout_10 = QGridLayout(self.loopstep_configuration_widget)
        self.gridLayout_10.setObjectName(u"gridLayout_10")
        self.gridLayout_10.setContentsMargins(0, 0, 0, 0)

        self.gridLayout_9.addWidget(self.loopstep_configuration_widget, 1, 0, 1, 1)

        self.meas_splitter.addWidget(self.configuration_main_widget)

        self.gridLayout.addWidget(self.meas_splitter, 0, 0, 1, 1)


        self.retranslateUi(Protocol_View)

        QMetaObject.connectSlotsByName(Protocol_View)
    # setupUi

    def retranslateUi(self, Protocol_View):
        Protocol_View.setWindowTitle(QCoreApplication.translate("Protocol_View", u"Form", None))
        self.pushButton_remove_step.setText(QCoreApplication.translate("Protocol_View", u"-", None))
        self.toolButton_add_step.setText(QCoreApplication.translate("Protocol_View", u"+", None))
        self.label_sequence.setText(QCoreApplication.translate("Protocol_View", u"Sequence", None))
        self.pushButton_move_step_up.setText(QCoreApplication.translate("Protocol_View", u"move up \u2191", None))
        self.pushButton_move_step_down.setText(QCoreApplication.translate("Protocol_View", u"move down \u2193", None))
        self.pushButton_move_step_out.setText(QCoreApplication.translate("Protocol_View", u"move out \u2190", None))
        self.pushButton_move_step_in.setText(QCoreApplication.translate("Protocol_View", u"move in \u2192", None))
        self.label_configuration.setText(QCoreApplication.translate("Protocol_View", u"Configuration", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'instrument_config.ui'
##
## Created by: Qt User Interface Compiler version 6.5.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QGridLayout,
    QHeaderView,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QTabWidget,
    QTableWidget,
    QTableWidgetItem,
    QWidget,
)


class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName("Form")
        Form.resize(483, 231)
        self.gridLayout = QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.gridLayout.setContentsMargins(9, 9, 9, 9)
        self.tableWidget_instruments = QTableWidget(Form)
        self.tableWidget_instruments.setObjectName("tableWidget_instruments")

        self.gridLayout.addWidget(self.tableWidget_instruments, 1, 0, 2, 2)

        self.label_config = QLabel(Form)
        self.label_config.setObjectName("label_config")
        font = QFont()
        font.setPointSize(10)
        font.setBold(True)
        self.label_config.setFont(font)

        self.gridLayout.addWidget(self.label_config, 0, 3, 1, 1)

        self.pushButton_remove = QPushButton(Form)
        self.pushButton_remove.setObjectName("pushButton_remove")
        self.pushButton_remove.setEnabled(False)
        self.pushButton_remove.setMaximumSize(QSize(89999, 27))
        font1 = QFont()
        font1.setBold(True)
        font1.setUnderline(False)
        font1.setStrikeOut(False)
        self.pushButton_remove.setFont(font1)
        self.pushButton_remove.setStyleSheet(
            "QPushButton {\n"
            "                                background-color:  #FF3333; \n"
            "                                color: white; \n"
            "                                border: none; \n"
            "                                padding: 0px; \n"
            "                                padding-bottom: 5px;\n"
            "                                text-align: center; \n"
            "                                text-decoration: none; \n"
            "                                font-size: 18px; \n"
            "                                margin: 2px 2px; \n"
            "                                border-radius: 6px;\n"
            "								font-weight: bold;\n"
            "                            }\n"
            "\n"
            "            QPushButton:hover {\n"
            "                background-color: #B22222;\n"
            "            }\n"
            "QPushButton:disabled {\n"
            "        background-color:#808080;\n"
            "    }"
        )

        self.gridLayout.addWidget(self.pushButton_remove, 1, 4, 1, 1)

        self.pushButton_info = QPushButton(Form)
        self.pushButton_info.setObjectName("pushButton_info")
        self.pushButton_info.setStyleSheet(
            "QPushButton {\n"
            "    background-color: #2a4cdf; /* This is a lighter shade of blue */\n"
            "    color: white;\n"
            "    border: none;\n"
            "    padding: 2px 10px;\n"
            "    text-align: center;\n"
            "    text-decoration: none;\n"
            "    font-size: 13px;\n"
            "    margin: 2px 2px;\n"
            "    border-radius: 6px;\n"
            "    font-weight: bold;\n"
            "}\n"
            "\n"
            "QPushButton:hover {\n"
            "    background-color: #1a3cbf; /* This is a darker shade of the lighter blue for hover effect */\n"
            "}"
        )

        self.gridLayout.addWidget(self.pushButton_info, 0, 4, 1, 1)

        self.label = QLabel(Form)
        self.label.setObjectName("label")

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.lineEdit_search = QLineEdit(Form)
        self.lineEdit_search.setObjectName("lineEdit_search")

        self.gridLayout.addWidget(self.lineEdit_search, 0, 1, 1, 1)

        self.config_tabs = QTabWidget(Form)
        self.config_tabs.setObjectName("config_tabs")
        self.tab = QWidget()
        self.tab.setObjectName("tab")
        self.gridLayout_2 = QGridLayout(self.tab)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.config_widge = QLabel(self.tab)
        self.config_widge.setObjectName("config_widge")

        self.gridLayout_2.addWidget(self.config_widge, 0, 0, 1, 1)

        self.config_tabs.addTab(self.tab, "")

        self.gridLayout.addWidget(self.config_tabs, 2, 3, 1, 2)

        self.pushButton_add = QPushButton(Form)
        self.pushButton_add.setObjectName("pushButton_add")
        self.pushButton_add.setEnabled(False)
        self.pushButton_add.setMaximumSize(QSize(99999, 27))
        self.pushButton_add.setFont(font1)
        self.pushButton_add.setStyleSheet(
            "QPushButton {\n"
            "                                background-color: #4CAF50; \n"
            "                                color: white; \n"
            "                                border: none; \n"
            "                                padding: 0px; \n"
            "                                padding-bottom: 5px;\n"
            "                                text-align: center; \n"
            "                                text-decoration: none; \n"
            "                                font-size: 18px; \n"
            "                                margin: 2px 2px; \n"
            "                                border-radius: 6px;\n"
            "								font-weight: bold;\n"
            "                            }\n"
            "\n"
            "                            QPushButton:hover {\n"
            "                                background-color: #45a049;\n"
            "                            }\n"
            "QPushButton:disabled {\n"
            "        background-color: #808080;\n"
            "    }"
        )

        self.gridLayout.addWidget(self.pushButton_add, 1, 3, 1, 1)

        self.retranslateUi(Form)

        self.config_tabs.setCurrentIndex(0)

        QMetaObject.connectSlotsByName(Form)

    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", "Form", None))
        self.label_config.setText(
            QCoreApplication.translate("Form", "Configure:", None)
        )
        self.pushButton_remove.setText(QCoreApplication.translate("Form", "-", None))
        self.pushButton_info.setText(
            QCoreApplication.translate("Form", "hide info", None)
        )
        self.label.setText(QCoreApplication.translate("Form", "Search name:", None))
        self.lineEdit_search.setPlaceholderText(
            QCoreApplication.translate("Form", "Search", None)
        )
        self.config_widge.setText(
            QCoreApplication.translate(
                "Form",
                '<html><head/><body><p><span style=" font-size:10pt; font-weight:600;">Select an instrument</span></p></body></html>',
                None,
            )
        )
        self.config_tabs.setTabText(
            self.config_tabs.indexOf(self.tab),
            QCoreApplication.translate("Form", "Select an instrument", None),
        )
        self.pushButton_add.setText(QCoreApplication.translate("Form", "+", None))

    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'set_channels.ui'
##
## Created by: Qt User Interface Compiler version 6.5.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import QApplication, QSizePolicy, QWidget


class Ui_Form(object):
    """ """

    def setupUi(self, Form):
        """

        Parameters
        ----------
        Form :


        Returns
        -------

        """
        if not Form.objectName():
            Form.setObjectName("Form")
        Form.resize(400, 300)

        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)

    # setupUi

    def retranslateUi(self, Form):
        """

        Parameters
        ----------
        Form :


        Returns
        -------

        """
        Form.setWindowTitle(QCoreApplication.translate("Form", "Form", None))

    # retranslateUi



from ophyd import Device, Kind
import bluesky.plan_stubs as bps


def trigger_and_read_devices(devices, name="primary"):
    """A wrapper for the plan stup `trigger_and_read` that implements
    trigger/read for a device with several components by simply splitting
    up the device and using a list of its components.

    Parameters
    ----------
    devices : list[ophyd.Device]
        List of the devices that should be triggered and read
    name : str
        (Default value = 'primary')
        Name of the stream where to save the data
    """
    split_devices = []
    for dev in devices:
        if isinstance(dev, Device):
            dev.read()
            for _, component in dev._get_components_of_kind(Kind.normal):
                split_devices.append(component)
        else:
            split_devices.append(dev)
    return bps.trigger_and_read(split_devices, name)



import re
import subprocess
import importlib

import pkg_resources

from nomad_camels.gui.device_installer import Ui_Form
from PySide6.QtWidgets import (
    QWidget,
    QTableWidgetItem,
    QCheckBox,
    QMessageBox,
    QTextBrowser,
    QSplitter,
    QTextEdit,
)
from PySide6.QtGui import QBrush, QColor, QFont, QPainter
from PySide6.QtCore import Qt, QThread, Signal

import requests
import sys
import os

from nomad_camels.utility.variables_handling import get_color
from nomad_camels.utility import variables_handling, device_handling
from nomad_camels.ui_widgets.warn_popup import WarnPopup

if sys.version_info >= (3, 8):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata

all_instr = {}
installed_instr = {}
last_repo = ""
last_branch = ""
last_dir = ""
repo_url = "https://raw.githubusercontent.com/FAU-LAP/CAMELS_drivers/main"


def getInstalledDevices(force=False, return_packages=False):
    """Goes through the given device_driver_path and returns a list of
    the available devices.

    Parameters
    ----------
    force :
         (Default value = False)
    return_packages :
         (Default value = False)

    Returns
    -------

    """
    global installed_instr
    if installed_instr and not force and not return_packages:
        return installed_instr
    installed_instr.clear()
    for x in importlib_metadata.distributions():
        name = x.metadata["Name"]
        version = x.version
        camels_driver_regex = r"^(nomad[-_]{1}camels[-_]{1}driver[-_]{1})(.*)$"
        if re.match(camels_driver_regex, name):
            installed_instr[name[20:].replace("-", "_")] = version
    packages = dict(device_handling.load_local_packages())
    for package in packages:
        installed_instr[package] = "local"
    not_loaded = []
    for instr in installed_instr:
        if instr not in packages:
            try:
                packages[instr] = importlib.import_module(
                    f"nomad_camels_driver_{instr}.{instr}"
                )
            except Exception as e:
                print(instr, e)
                not_loaded.append(instr)
    for instr in not_loaded:
        installed_instr.pop(instr)
    if return_packages:
        return installed_instr, packages
    return installed_instr


def getAllDevices():
    """So far only returns the installed devices, should in future work
    with the online repository of drivers.

    Parameters
    ----------

    Returns
    -------

    """
    global all_instr, last_repo, last_branch, last_dir, repo_url
    try:
        repo = variables_handling.preferences["driver_repository"]
        branch = variables_handling.preferences["repo_branch"]
        directory = variables_handling.preferences["repo_directory"]
    except:
        repo, branch, directory = "", "", ""
    if (
        all_instr
        and last_repo == repo
        and branch == last_branch
        and directory == last_dir
    ):
        return all_instr
    last_repo, last_branch, last_dir = repo, branch, directory
    all_instr = {}
    try:
        # repo_part = repo.split(".com/")[1].split(".git")[0]
        # repo_url = f"https://raw.githubusercontent.com/{repo_part}/{branch}/{directory}"
        url = "https://raw.githubusercontent.com/FAU-LAP/CAMELS_drivers/driver_list/driver_list.txt"
        devices_str = requests.get(url).text
    except:
        try:
            url = "https://raw.githubusercontent.com/FAU-LAP/CAMELS_drivers/main/driver_list.txt"
            devices_str = requests.get(url).text
        except:
            devices_str = ""
    warned = False
    for x in devices_str.splitlines():
        if "==" not in x:
            continue
        try:
            name, version = x.split("==")
        except:
            if not warned:
                WarnPopup(
                    None,
                    "Could not (completely) read driver_list.txt from repo.\n"
                    "Check settings if repository and branch are correct.",
                    "No online-drivers found",
                )
            warned = True
            continue
        all_instr[name.replace("-", "_")] = version
    return all_instr


def get_instr_readme_text(instr_name):
    url = f"{repo_url}/{instr_name}/README.md"
    return requests.get(url).text


def get_instr_license_text(instr_name):
    url = f"{repo_url}/{instr_name}/LICENSE.txt"
    return requests.get(url).text


bold_font = QFont()
bold_font.setBold(True)


class Info_Widget(QSplitter):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.splitter = QSplitter(Qt.Orientation.Vertical)

        self.info_text = QTextBrowser()
        self.info_text.setOpenExternalLinks(True)
        self.info_text.setTextInteractionFlags(
            Qt.TextSelectableByKeyboard
            | Qt.TextSelectableByMouse
            | Qt.LinksAccessibleByMouse
        )
        self.addWidget(self.info_text)

        self.license_text = QTextBrowser()
        self.license_text.setOpenExternalLinks(True)
        self.license_text.setTextInteractionFlags(
            Qt.TextSelectableByKeyboard
            | Qt.TextSelectableByMouse
            | Qt.LinksAccessibleByMouse
        )
        self.addWidget(self.license_text)
        self.info = False

        self.setContentsMargins(0, 0, 0, 0)

    def update_texts(self, instr):
        self.info = False
        hidden = True
        try:
            self.info_text.setHidden(True)
            self.info_text.clear()
            try:
                meta = importlib_metadata.metadata(f"nomad_camels_driver_{instr}")
                text = meta.json["description"]
            except:
                text = get_instr_readme_text(instr)
            self.info_text.setMarkdown(text)
            self.info_text.setHidden(False)
            hidden = False
        except:
            pass
        try:
            self.license_text.setHidden(True)
            self.license_text.clear()
            try:
                text = ""
                for p in pkg_resources.working_set:
                    if not p.key.startswith(
                        f'nomad-camels-driver-{instr.replace("_", "-")}'
                    ):
                        continue
                    lic = p.get_metadata_lines("LICENSE.txt")
                    for l in lic:
                        text += f"{l}\n"
                    break
                if not text:
                    raise Exception("")
                self.license_text.setText(text)
            except:
                text = get_instr_license_text(instr)
                self.license_text.setMarkdown(text)
            self.license_text.setHidden(False)
            hidden = False
        except:
            pass
        self.info = not hidden
        self.setHidden(hidden)


class Instrument_Installer(Ui_Form, QWidget):
    """ """

    instruments_updated = Signal()

    """
    Parameters
    ----------
    parent : QWidget
        handed over to QWidget."""

    def __init__(self, active_instruments=None, parent=None):
        super().__init__(parent=parent)
        self.setupUi(self)
        # Overwrite the existing TextEdit with one that writes gray background text if there is nothing in the installation log
        textEdit_device_info = CustomTextEdit_new_painter()
        self.layout().replaceWidget(self.textEdit_device_info, textEdit_device_info)
        self.textEdit_device_info.setParent(None)
        self.textEdit_device_info.deleteLater()
        self.textEdit_device_info = textEdit_device_info
        self.show_info_bool = False
        self.active_instruments = active_instruments or {}

        self.device_table.setColumnCount(3)
        self.device_table.setHorizontalHeaderLabels(
            ["instrument", "available", "installed"]
        )
        self.device_table.verticalHeader().setVisible(False)

        self.all_devs = getAllDevices()
        self.installed_devs = getInstalledDevices()
        self.checkboxes = []

        self.disables = []
        for c in self.children():
            if c not in [
                self.progressBar,
                self.textEdit_device_info,
                self.label,
                self.label_2,
            ]:
                self.disables.append(c)

        self.pushButton_sel_all.clicked.connect(self.select_all)
        self.pushButton_sel_none.clicked.connect(self.select_none)
        self.lineEdit_search_name.textChanged.connect(self.build_table)
        self.build_table()

        self.install_thread = None

        self.pushButton_install_update_selected.clicked.connect(self.install_selected)
        self.pushButton_uninstall.clicked.connect(self.uninstall_selected)
        self.pushButton_update_drivers.clicked.connect(self.update_installed)
        self.pushButton_info.clicked.connect(self.show_hide_info)

        self.device_table.clicked.connect(self.table_click)

        self.info_widge = Info_Widget()
        self.info_widge.setHidden(True)
        self.layout().addWidget(self.info_widge, 0, 5, 8, 1)
        self.info_widge.setHidden(True)

    def show_hide_info(self):
        if self.info_widge.isHidden():
            self.info_widge.setHidden(False)
            self.adjustSize()
            self.pushButton_info.setText("hide info")
            self.show_info_bool = True
        else:
            self.info_widge.setHidden(True)
            self.adjustSize()
            self.pushButton_info.setText("show info")
            self.show_info_bool = False

    def checkBox_change(self, row):
        """

        Parameters
        ----------
        row :


        Returns
        -------

        """
        for i in range(3):
            checked = self.device_table.cellWidget(row, 0).isChecked()
            item = self.device_table.item(row, i)
            brush = QBrush(QColor(get_color("blue" if checked else "white")))
            item.setBackground(brush)

    def get_checked_devs(self, ignore_version=False):
        """

        Parameters
        ----------
        ignore_version :
             (Default value = False)

        Returns
        -------

        """
        devs = []
        for box in self.checkboxes:
            if not box.isChecked():
                continue
            dev = box.text()
            if ignore_version:
                remove_dialog = QMessageBox.question(
                    self,
                    "Uninstall instrument?",
                    f'You are trying to uninstall the instrument "{dev}", but it may still be in use.\nContinue?',
                    QMessageBox.Yes | QMessageBox.No,
                )
                if remove_dialog != QMessageBox.Yes:
                    continue
                devs.append(dev)
                continue
            if (
                dev in self.installed_devs
                and self.all_devs[dev] == self.installed_devs[dev]
            ):
                continue
            devs.append(dev)
        return devs

    def update_installed(self):
        """ """
        devs = []
        for dev, version in self.installed_devs.items():
            if version != self.all_devs[dev]:
                devs.append(dev)
        self.install_thread = Install_Thread(devs, False, self)
        self.install_thread.info_step.connect(self.textEdit_device_info.append)
        self.install_thread.val_step.connect(self.progressBar.setValue)
        self.install_thread.finished.connect(self.thread_done)
        for c in self.disables:
            c.setEnabled(False)
        self.install_thread.start()

    def uninstall_selected(self):
        """ """
        self.install_selected(True)

    def install_selected(self, uninstall=False):
        """

        Parameters
        ----------
        uninstall :
             (Default value = False)

        Returns
        -------

        """
        self.textEdit_device_info.clear()
        devs = self.get_checked_devs(uninstall)
        self.install_thread = Install_Thread(devs, uninstall, self)
        self.install_thread.info_step.connect(self.textEdit_device_info.append)
        self.install_thread.val_step.connect(self.progressBar.setValue)
        self.install_thread.finished.connect(self.thread_done)
        for c in self.disables:
            c.setEnabled(False)
        self.setCursor(Qt.WaitCursor)
        self.install_thread.start()

    def thread_done(self):
        """ """
        for c in self.disables:
            c.setEnabled(True)
        self.setCursor(Qt.ArrowCursor)
        self.all_devs = getAllDevices()
        self.installed_devs = getInstalledDevices(True)
        self.build_table()
        self.instruments_updated.emit()

    def select_all(self):
        """ """
        for box in self.checkboxes:
            box.setChecked(True)

    def select_none(self):
        """ """
        for box in self.checkboxes:
            box.setChecked(False)

    def build_table(self):
        """ """
        search_text = self.lineEdit_search_name.text()
        self.checkboxes.clear()
        self.device_table.clear()
        self.device_table.setRowCount(0)
        self.device_table.setHorizontalHeaderLabels(
            ["instrument", "available", "installed"]
        )
        i = 0
        for dev in sorted(self.all_devs.keys()):
            if search_text.lower() not in dev.lower():
                continue
            self.device_table.setRowCount(i + 1)
            item = QTableWidgetItem()
            item.setData(3, dev)
            checkbox = QCheckBox(dev)
            self.checkboxes.append(checkbox)
            checkbox.stateChanged.connect(lambda a0, x=i: self.checkBox_change(x))
            self.device_table.setCellWidget(i, 0, checkbox)

            item_v = QTableWidgetItem(self.all_devs[dev])
            item_v.setFont(bold_font)
            item_v.setFlags(item_v.flags() & ~Qt.ItemIsEditable)
            inst_v = self.installed_devs[dev] if dev in self.installed_devs else ""
            item_inst = QTableWidgetItem(inst_v)
            if inst_v:
                brush = QBrush(
                    QColor(
                        get_color(
                            "dark_green" if inst_v == self.all_devs[dev] else "orange"
                        )
                    )
                )
                item_inst.setForeground(brush)
            item_inst.setFlags(item_inst.flags() & ~Qt.ItemIsEditable)
            item_inst.setFont(bold_font)
            self.device_table.setItem(i, 0, item)
            self.device_table.setItem(i, 1, item_v)
            self.device_table.setItem(i, 2, item_inst)
            i += 1

    def table_click(self):
        """ """
        self.setCursor(Qt.WaitCursor)
        try:
            ind = self.device_table.selectedIndexes()[0]
            instr = self.device_table.item(ind.row(), 0).data(3)
            self.info_widge.update_texts(instr)
            self.adjustSize()
            if self.show_info_bool:
                self.info_widge.setHidden(False)
            else:
                self.info_widge.setHidden(True)
        finally:
            self.setCursor(Qt.ArrowCursor)


class CustomTextEdit_new_painter(QTextEdit):
    """Custom QTextEdit that paints custom text when the text is empty."""

    def paintEvent(self, event):
        # Call the original paint event to handle the default drawing
        super().paintEvent(event)

        # Custom painting code
        if not self.toPlainText():
            painter = QPainter(self.viewport())
            painter.setPen(QColor(Qt.gray))  # Set the color of the pen

            # Customize the font
            font = QFont()
            font.setPointSize(16)  # Adjust the size as needed
            font.setBold(True)
            painter.setFont(font)

            # Draw custom text in the center
            painter.drawText(self.rect(), Qt.AlignCenter, "Installation Log")
            painter.end()


class Install_Thread(QThread):
    """ """

    info_step = Signal(str)
    val_step = Signal(int)

    def __init__(self, devs, uninstall=False, parent=None):
        super().__init__(parent=parent)
        self.devs = devs
        self.uninstall = uninstall

    def run(self):
        """ """
        n = len(self.devs)
        for i, dev in enumerate(self.devs):
            self.val_step.emit(i / n * 100)
            if self.uninstall:
                device_name = dev.replace("_", "-")
                flags = 0
                if os.name == "nt":
                    flags = subprocess.CREATE_NO_WINDOW
                ret = subprocess.Popen(
                    [
                        sys.executable,
                        "-m",
                        "pip",
                        "uninstall",
                        "-y",
                        f"nomad-camels-driver-{device_name}",
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    stdin=subprocess.PIPE,
                    creationflags=flags,
                )
                if ret.returncode:
                    raise OSError(
                        f"Failed to uninstall nomad-camels-driver-{device_name}"
                    )
            else:
                device_name = dev.replace("_", "-")
                ret = install_instrument(device_name)
            for line in iter(ret.stdout.readline, b""):
                text = line.decode().rstrip()
                self.info_step.emit(text)
        getInstalledDevices(True)
        for i, dev in enumerate(self.devs):
            if self.uninstall and dev in installed_instr:
                raise Warning(f"Uninstall of {dev} failed!")
                # WarnPopup(self.parent(),
                #           f'Uninstall of {dev} failed!',
                #           f'Uninstall of {dev} failed!')
            elif not self.uninstall and dev not in installed_instr:
                raise Warning(f"Installation of {dev} failed!")
                # WarnPopup(self.parent(),
                #           f'Installation of {dev} failed!',
                #           f'Installation of {dev} failed!')
        self.val_step.emit(100)


def install_instrument(device_name):
    """

    Parameters
    ----------
    device_name :


    Returns
    -------

    """
    # pypi_url = r'https://test.pypi.org/simple/'  # TODO Change to regular PyPi
    flags = 0
    if os.name == "nt":
        flags = subprocess.CREATE_NO_WINDOW
    ret = subprocess.Popen(
        [
            sys.executable,
            "-m",
            "pip",
            "install",
            # '--no-cache-dir',
            # '--index-url', pypi_url,
            # '--extra-index-url',
            # 'https://pypi.org/simple',
            f"nomad-camels-driver-{device_name}",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        stdin=subprocess.PIPE,
        creationflags=flags,
    )
    if ret.returncode:
        raise OSError(f"Failed to install nomad-camels-driver-{device_name}")
    return ret


if __name__ == "__main__":
    from PySide6.QtWidgets import QApplication

    app = QApplication([])

    widge = Instrument_Installer()
    widge.show()
    app.exec()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'plot_definer_2d.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QGridLayout, QLabel, QLineEdit,
    QSizePolicy, QSpacerItem, QWidget)

from nomad_camels.ui_widgets.variable_tool_tip_box import Variable_Box

class Ui_Plot_Definer_2D(object):
    def setupUi(self, Plot_Definer_2D):
        if not Plot_Definer_2D.objectName():
            Plot_Definer_2D.setObjectName(u"Plot_Definer_2D")
        Plot_Definer_2D.resize(400, 222)
        self.gridLayout = QGridLayout(Plot_Definer_2D)
        self.gridLayout.setObjectName(u"gridLayout")
        self.label_7 = QLabel(Plot_Definer_2D)
        self.label_7.setObjectName(u"label_7")
        font = QFont()
        font.setPointSize(9)
        font.setBold(True)
        self.label_7.setFont(font)

        self.gridLayout.addWidget(self.label_7, 4, 0, 1, 1)

        self.label_6 = QLabel(Plot_Definer_2D)
        self.label_6.setObjectName(u"label_6")
        self.label_6.setFont(font)

        self.gridLayout.addWidget(self.label_6, 3, 0, 1, 1)

        self.label_8 = QLabel(Plot_Definer_2D)
        self.label_8.setObjectName(u"label_8")
        self.label_8.setFont(font)

        self.gridLayout.addWidget(self.label_8, 3, 2, 1, 1)

        self.lineEdit_ylabel = QLineEdit(Plot_Definer_2D)
        self.lineEdit_ylabel.setObjectName(u"lineEdit_ylabel")

        self.gridLayout.addWidget(self.lineEdit_ylabel, 3, 3, 1, 1)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer_2, 5, 1, 1, 1)

        self.label_10 = QLabel(Plot_Definer_2D)
        self.label_10.setObjectName(u"label_10")
        self.label_10.setFont(font)

        self.gridLayout.addWidget(self.label_10, 4, 2, 1, 1)

        self.label_9 = QLabel(Plot_Definer_2D)
        self.label_9.setObjectName(u"label_9")
        self.label_9.setFont(font)

        self.gridLayout.addWidget(self.label_9, 2, 2, 1, 1)

        self.label_5 = QLabel(Plot_Definer_2D)
        self.label_5.setObjectName(u"label_5")
        self.label_5.setFont(font)

        self.gridLayout.addWidget(self.label_5, 2, 0, 1, 1)

        self.lineEdit_title = QLineEdit(Plot_Definer_2D)
        self.lineEdit_title.setObjectName(u"lineEdit_title")

        self.gridLayout.addWidget(self.lineEdit_title, 0, 1, 1, 3)

        self.label_11 = QLabel(Plot_Definer_2D)
        self.label_11.setObjectName(u"label_11")
        self.label_11.setFont(font)

        self.gridLayout.addWidget(self.label_11, 0, 0, 1, 1)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 1, 1, 1, 1)

        self.lineEdit_y_axis = Variable_Box(Plot_Definer_2D)
        self.lineEdit_y_axis.setObjectName(u"lineEdit_y_axis")

        self.gridLayout.addWidget(self.lineEdit_y_axis, 3, 1, 1, 1)

        self.lineEdit_z_axis = Variable_Box(Plot_Definer_2D)
        self.lineEdit_z_axis.setObjectName(u"lineEdit_z_axis")

        self.gridLayout.addWidget(self.lineEdit_z_axis, 4, 1, 1, 1)

        self.lineEdit_xlabel = QLineEdit(Plot_Definer_2D)
        self.lineEdit_xlabel.setObjectName(u"lineEdit_xlabel")

        self.gridLayout.addWidget(self.lineEdit_xlabel, 2, 3, 1, 1)

        self.lineEdit_x_axis = Variable_Box(Plot_Definer_2D)
        self.lineEdit_x_axis.setObjectName(u"lineEdit_x_axis")

        self.gridLayout.addWidget(self.lineEdit_x_axis, 2, 1, 1, 1)

        self.lineEdit_zlabel = QLineEdit(Plot_Definer_2D)
        self.lineEdit_zlabel.setObjectName(u"lineEdit_zlabel")

        self.gridLayout.addWidget(self.lineEdit_zlabel, 4, 3, 1, 1)

        self.label_12 = QLabel(Plot_Definer_2D)
        self.label_12.setObjectName(u"label_12")
        self.label_12.setFont(font)

        self.gridLayout.addWidget(self.label_12, 6, 0, 1, 2)

        self.lineEdit_n_data_points = QLineEdit(Plot_Definer_2D)
        self.lineEdit_n_data_points.setObjectName(u"lineEdit_n_data_points")

        self.gridLayout.addWidget(self.lineEdit_n_data_points, 6, 2, 1, 2)

        QWidget.setTabOrder(self.lineEdit_title, self.lineEdit_x_axis)
        QWidget.setTabOrder(self.lineEdit_x_axis, self.lineEdit_y_axis)
        QWidget.setTabOrder(self.lineEdit_y_axis, self.lineEdit_z_axis)
        QWidget.setTabOrder(self.lineEdit_z_axis, self.lineEdit_xlabel)
        QWidget.setTabOrder(self.lineEdit_xlabel, self.lineEdit_ylabel)
        QWidget.setTabOrder(self.lineEdit_ylabel, self.lineEdit_zlabel)

        self.retranslateUi(Plot_Definer_2D)

        QMetaObject.connectSlotsByName(Plot_Definer_2D)
    # setupUi

    def retranslateUi(self, Plot_Definer_2D):
        Plot_Definer_2D.setWindowTitle(QCoreApplication.translate("Plot_Definer_2D", u"Form", None))
        self.label_7.setText(QCoreApplication.translate("Plot_Definer_2D", u"z-axis:", None))
        self.label_6.setText(QCoreApplication.translate("Plot_Definer_2D", u"y-axis:", None))
        self.label_8.setText(QCoreApplication.translate("Plot_Definer_2D", u"y-label:", None))
        self.label_10.setText(QCoreApplication.translate("Plot_Definer_2D", u"z-label:", None))
        self.label_9.setText(QCoreApplication.translate("Plot_Definer_2D", u"x-label:", None))
        self.label_5.setText(QCoreApplication.translate("Plot_Definer_2D", u"x-axis:", None))
        self.label_11.setText(QCoreApplication.translate("Plot_Definer_2D", u"title:", None))
        self.label_12.setText(QCoreApplication.translate("Plot_Definer_2D", u"# data points:", None))
        self.lineEdit_n_data_points.setText(QCoreApplication.translate("Plot_Definer_2D", u"inf", None))
    # retranslateUi




import pytest
from PySide6.QtGui import QStandardItemModel
from PySide6.QtWidgets import QApplication

from nomad_camels.main_classes.loop_step import (
    Loop_Step,
    Loop_Step_Container,
    Loop_Step_Config,
    Loop_Step_Name_Widget,
)


# Initialize the Qt Application for testing


@pytest.fixture(scope="module")
def app():
    """Fixture for creating a QApplication instance"""
    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    return app


def test_loop_step_init():
    step = Loop_Step(
        name="Test Step",
        step_info={"description": "Test description", "is_active": True},
    )
    assert step.name == "Test Step"
    assert step.description == "Test description"
    assert step.is_active is True
    assert step.step_type == "Default"
    assert step.full_name == "Default (Test Step)"
    assert step.has_children is False
    assert step.children == []


def test_loop_step_update_full_name():
    step = Loop_Step(name="Test Step")
    step.update_full_name()
    assert step.full_name == "Default (Test Step)"


def test_loop_step_append_to_model(app):
    step = Loop_Step(name="Test Step")
    model = QStandardItemModel()
    item = step.append_to_model(model)
    assert item.text() == "Default (Test Step)"


def test_loop_step_get_protocol_string():
    step = Loop_Step(name="Test Step", step_info={"description": "Test description"})
    protocol_string = step.get_protocol_string()
    assert "Test description" in protocol_string


def test_loop_step_container_init():
    step = Loop_Step_Container(name="Container Step")
    assert step.name == "Container Step"
    assert step.step_type == "Container"
    assert step.has_children is True
    assert step.children == []


def test_loop_step_container_add_child():
    parent_step = Loop_Step_Container(name="Parent Step")
    child_step = Loop_Step(name="Child Step")
    parent_step.add_child(child_step)
    assert len(parent_step.children) == 1
    assert parent_step.children[0] == child_step


def test_loop_step_container_get_protocol_string():
    parent_step = Loop_Step_Container(name="Parent Step")
    child_step = Loop_Step(
        name="Child Step", step_info={"description": "Child description"}
    )
    parent_step.add_child(child_step)
    parent_step.update_time_weight()
    assert parent_step.time_weight == 2


def test_loop_step_name_widget_init(app):
    widget = Loop_Step_Name_Widget(name="Test Step", is_active=True)
    assert widget.lineEdit_name.text() == "Test Step"
    assert widget.checkBox_active.isChecked() is True


def test_loop_step_config_init(app):
    step = Loop_Step(
        name="Test Step",
        step_info={"description": "Test description", "is_active": True},
    )
    config_widget = Loop_Step_Config(loop_step=step)
    assert config_widget.name_widget.lineEdit_name.text() == "Test Step"
    assert config_widget.textEdit_desc.toPlainText() == "Test description"


def test_loop_step_config_change_name(app):
    step = Loop_Step(name="Old Name")
    config_widget = Loop_Step_Config(loop_step=step)
    config_widget.change_name("New Name")
    assert step.name == "New Name"
    assert step.full_name == "Default (New Name)"


def test_loop_step_config_change_active(app):
    step = Loop_Step(is_active=True)
    config_widget = Loop_Step_Config(loop_step=step)
    config_widget.change_active(False)
    assert step.is_active is False


def test_loop_step_config_update_step_config(app):
    step = Loop_Step(name="Test_Step", step_info={"description": "Old description"})
    config_widget = Loop_Step_Config(loop_step=step)
    config_widget.textEdit_desc.setPlainText("New description")
    config_widget.update_step_config()
    assert step.description == "New description"



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'device_installer.ui'
##
## Created by: Qt User Interface Compiler version 6.5.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QGridLayout,
    QHeaderView,
    QLabel,
    QLineEdit,
    QProgressBar,
    QPushButton,
    QSizePolicy,
    QTableWidget,
    QTableWidgetItem,
    QTextEdit,
    QWidget,
)
class CustomTextEdit(QTextEdit):
    def __init__(self, parent=None):
        super().__init__(parent)

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName("Form")
        Form.resize(679, 481)
        self.gridLayout = QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.textEdit_device_info = CustomTextEdit(Form)
        self.textEdit_device_info.setObjectName("textEdit_device_info")
        self.textEdit_device_info.setEnabled(True)
        self.textEdit_device_info.setTextInteractionFlags(
            Qt.TextSelectableByKeyboard | Qt.TextSelectableByMouse
        )

        self.gridLayout.addWidget(self.textEdit_device_info, 1, 2, 5, 1)

        self.pushButton_update_drivers = QPushButton(Form)
        self.pushButton_update_drivers.setObjectName("pushButton_update_drivers")
        self.pushButton_update_drivers.setEnabled(True)
        self.pushButton_update_drivers.setStyleSheet(
            "QPushButton {\n"
            "                                background-color: #2a4cdf; \n"
            "                                color: white; \n"
            "                                border: none; \n"
            "                                padding: 2px 10px; \n"
            "                                text-align: center; \n"
            "                                text-decoration: none; \n"
            "                                font-size: 13px; \n"
            "                                margin: 2px 2px; \n"
            "                                border-radius: 6px;\n"
            "								font-weight: bold;\n"
            "                            }\n"
            "\n"
            "                            QPushButton:hover {\n"
            "                                background-color: #1a3cbf;\n"
            "                            }"
        )

        self.gridLayout.addWidget(self.pushButton_update_drivers, 7, 0, 1, 1)

        self.label = QLabel(Form)
        self.label.setObjectName("label")

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.pushButton_sel_none = QPushButton(Form)
        self.pushButton_sel_none.setObjectName("pushButton_sel_none")
        self.pushButton_sel_none.setStyleSheet(
            "QPushButton {\n"
            "        background-color: #A9A9A9;\n"
            "        color: white;\n"
            "        border: none;\n"
            "        padding: 2px 10px;\n"
            "        text-align: center;\n"
            "        text-decoration: none;\n"
            "        font-size: 13px;\n"
            "        margin: 2px 2px;\n"
            "        border-radius: 6px;\n"
            "        font-weight: bold;\n"
            "    }\n"
            "\n"
            "    QPushButton:hover {\n"
            "        background-color: #696969;\n"
            "    }"
        )

        self.gridLayout.addWidget(self.pushButton_sel_none, 2, 1, 1, 1)

        self.pushButton_uninstall = QPushButton(Form)
        self.pushButton_uninstall.setObjectName("pushButton_uninstall")
        self.pushButton_uninstall.setEnabled(True)
        self.pushButton_uninstall.setStyleSheet(
            "QPushButton {\n"
            "                background-color: #E60000; \n"
            "                color: white; \n"
            "                                border: none; \n"
            "                                padding: 2px 10px; \n"
            "                                text-align: center; \n"
            "                                text-decoration: none; \n"
            "                                font-size: 13px; \n"
            "                                margin: 2px 2px; \n"
            "                                border-radius: 6px;\n"
            "								font-weight: bold;\n"
            "                            }\n"
            "\n"
            "            QPushButton:hover {\n"
            "                background-color: #B22222;\n"
            "            }"
        )

        self.gridLayout.addWidget(self.pushButton_uninstall, 6, 1, 1, 1)

        self.pushButton_install_update_selected = QPushButton(Form)
        self.pushButton_install_update_selected.setObjectName(
            "pushButton_install_update_selected"
        )
        self.pushButton_install_update_selected.setEnabled(True)
        self.pushButton_install_update_selected.setStyleSheet(
            "QPushButton {\n"
            "                                background-color: #4CAF50; \n"
            "                                color: white; \n"
            "                                border: none; \n"
            "                                padding: 2px 10px; \n"
            "                                text-align: center; \n"
            "                                text-decoration: none; \n"
            "                                font-size: 13px; \n"
            "                                margin: 2px 2px; \n"
            "                                border-radius: 6px;\n"
            "								font-weight: bold;\n"
            "                            }\n"
            "\n"
            "                            QPushButton:hover {\n"
            "                                background-color: #45a049;\n"
            "                            }"
        )

        self.gridLayout.addWidget(self.pushButton_install_update_selected, 6, 0, 1, 1)

        self.lineEdit_search_tags = QLineEdit(Form)
        self.lineEdit_search_tags.setObjectName("lineEdit_search_tags")
        self.lineEdit_search_tags.setMaximumSize(QSize(250, 16777215))

        self.gridLayout.addWidget(self.lineEdit_search_tags, 1, 1, 1, 1)

        self.pushButton_info = QPushButton(Form)
        self.pushButton_info.setObjectName("pushButton_info")
        self.pushButton_info.setStyleSheet("")

        self.gridLayout.addWidget(self.pushButton_info, 6, 2, 1, 1)

        self.device_table = QTableWidget(Form)
        self.device_table.setObjectName("device_table")

        self.gridLayout.addWidget(self.device_table, 3, 0, 3, 2)

        self.pushButton_sel_all = QPushButton(Form)
        self.pushButton_sel_all.setObjectName("pushButton_sel_all")
        self.pushButton_sel_all.setStyleSheet(
            "QPushButton {\n"
            "    background-color: #2a4cdf; /* This is a lighter shade of blue */\n"
            "    color: white;\n"
            "    border: none;\n"
            "    padding: 2px 10px;\n"
            "    text-align: center;\n"
            "    text-decoration: none;\n"
            "    font-size: 13px;\n"
            "    margin: 2px 2px;\n"
            "    border-radius: 6px;\n"
            "    font-weight: bold;\n"
            "}\n"
            "\n"
            "QPushButton:hover {\n"
            "    background-color: #1a3cbf; /* This is a darker shade of the lighter blue for hover effect */\n"
            "}"
        )

        self.gridLayout.addWidget(self.pushButton_sel_all, 2, 0, 1, 1)

        self.progressBar = QProgressBar(Form)
        self.progressBar.setObjectName("progressBar")
        self.progressBar.setEnabled(True)
        self.progressBar.setValue(0)
        self.progressBar.setInvertedAppearance(False)

        self.gridLayout.addWidget(self.progressBar, 0, 2, 1, 1)

        self.label_2 = QLabel(Form)
        self.label_2.setObjectName("label_2")

        self.gridLayout.addWidget(self.label_2, 1, 0, 1, 1)

        self.lineEdit_search_name = QLineEdit(Form)
        self.lineEdit_search_name.setObjectName("lineEdit_search_name")
        self.lineEdit_search_name.setMaximumSize(QSize(250, 16777215))

        self.gridLayout.addWidget(self.lineEdit_search_name, 0, 1, 1, 1)

        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)

    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", "Form", None))
        self.pushButton_update_drivers.setText(
            QCoreApplication.translate("Form", "Update all installed drivers", None)
        )
        self.label.setText(QCoreApplication.translate("Form", "Search name:", None))
        self.pushButton_sel_none.setText(
            QCoreApplication.translate("Form", "Select None", None)
        )
        self.pushButton_uninstall.setText(
            QCoreApplication.translate("Form", "Uninstall Selected", None)
        )
        self.pushButton_install_update_selected.setText(
            QCoreApplication.translate("Form", "Install / Update Selected", None)
        )
        self.lineEdit_search_tags.setPlaceholderText(
            QCoreApplication.translate("Form", "tags", None)
        )
        self.pushButton_info.setText(
            QCoreApplication.translate("Form", "show info", None)
        )
        self.pushButton_sel_all.setText(
            QCoreApplication.translate("Form", "Select All", None)
        )
        self.label_2.setText(QCoreApplication.translate("Form", "Search tags:", None))
        self.lineEdit_search_name.setPlaceholderText(
            QCoreApplication.translate("Form", "device name", None)
        )

    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'pass_ask.ui'
##
## Created by: Qt User Interface Compiler version 6.5.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QAbstractButton,
    QApplication,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QWidget,
)


class Ui_Pass_Ask(object):
    """ """

    def setupUi(self, Pass_Ask):
        """

        Parameters
        ----------
        Pass_Ask :


        Returns
        -------

        """
        if not Pass_Ask.objectName():
            Pass_Ask.setObjectName("Pass_Ask")
        Pass_Ask.resize(400, 106)
        self.gridLayout = QGridLayout(Pass_Ask)
        self.gridLayout.setObjectName("gridLayout")
        self.label_2 = QLabel(Pass_Ask)
        self.label_2.setObjectName("label_2")

        self.gridLayout.addWidget(self.label_2, 1, 0, 1, 1)

        self.lineEdit_password_1 = QLineEdit(Pass_Ask)
        self.lineEdit_password_1.setObjectName("lineEdit_password_1")
        self.lineEdit_password_1.setEchoMode(QLineEdit.Password)

        self.gridLayout.addWidget(self.lineEdit_password_1, 1, 1, 1, 1)

        self.buttonBox = QDialogButtonBox(Pass_Ask)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)

        self.gridLayout.addWidget(self.buttonBox, 2, 0, 1, 2)

        self.label = QLabel(Pass_Ask)
        self.label.setObjectName("label")

        self.gridLayout.addWidget(self.label, 0, 0, 1, 2)

        self.retranslateUi(Pass_Ask)
        self.buttonBox.accepted.connect(Pass_Ask.accept)
        self.buttonBox.rejected.connect(Pass_Ask.reject)

        QMetaObject.connectSlotsByName(Pass_Ask)

    # setupUi

    def retranslateUi(self, Pass_Ask):
        """

        Parameters
        ----------
        Pass_Ask :


        Returns
        -------

        """
        Pass_Ask.setWindowTitle(QCoreApplication.translate("Pass_Ask", "Dialog", None))
        self.label_2.setText(QCoreApplication.translate("Pass_Ask", "Password:", None))
        self.label.setText(
            QCoreApplication.translate(
                "Pass_Ask", 'Enter sudo-password for the Ubuntu-user "epics":', None
            )
        )

    # retranslateUi



from PySide6.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QTabWidget,
    QMessageBox,
)
from PySide6.QtCore import Qt
from PySide6.QtGui import QCloseEvent, QKeyEvent

from nomad_camels.frontpanels import instrument_installer, instrument_config


class ManageInstruments(QDialog):
    """ """

    def __init__(self, active_instruments=None, parent=None):
        super().__init__(parent=parent)
        self.buttonBox = QDialogButtonBox()
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.accepted.connect(self.accept)

        self.setWindowTitle("Manage Instruments - NOMAD CAMELS")
        self.setWindowFlags(self.windowFlags() | Qt.WindowMaximizeButtonHint)

        self.active_instruments = active_instruments or {}

        layout = QGridLayout()
        self.setLayout(layout)

        self.tabs = QTabWidget()

        self.installer = instrument_installer.Instrument_Installer()
        self.tabs.addTab(self.installer, "Install Instruments")

        self.config_widget = instrument_config.Instrument_Config(active_instruments)
        self.tabs.addTab(self.config_widget, "Configure Instruments")

        self.installer.instruments_updated.connect(self.config_widget.build_table)

        layout.addWidget(self.tabs, 0, 0)
        layout.addWidget(self.buttonBox, 1, 0)

        self.show()
        settab = 1 if instrument_installer.getInstalledDevices() else 0
        self.tabs.setCurrentIndex(settab)
        if not self.installer.all_devs:
            self.tabs.setTabEnabled(0, False)
            self.tabs.setTabToolTip(
                0,
                'Could not reach the server for list of drivers\ncheck your internet connection and try opening "Manage Instruments" again',
            )

    def accept(self) -> None:
        """ """
        self.active_instruments = self.config_widget.get_config()
        super().accept()

    def closeEvent(self, a0: QCloseEvent) -> None:
        """

        Parameters
        ----------
        a0: QCloseEvent :


        Returns
        -------

        """
        discard_dialog = QMessageBox.question(
            self,
            "Discard Changes?",
            f"All changes to instrument configurations will be lost!",
            QMessageBox.Yes | QMessageBox.No,
        )
        if discard_dialog != QMessageBox.Yes:
            a0.ignore()
            return
        super().closeEvent(a0)

    def keyPressEvent(self, a0: QKeyEvent) -> None:
        """Overwrites the keyPressEvent of the QDialog so that it does
        not close when pressing Enter/Return.

        Parameters
        ----------
        a0: QKeyEvent :


        Returns
        -------

        """
        if a0.key() == Qt.Key_Enter or a0.key() == Qt.Key_Return:
            return
        super().keyPressEvent(a0)


if __name__ == "__main__":
    from PySide6.QtWidgets import QApplication

    app = QApplication([])

    widge = ManageInstruments()
    widge.show()
    app.exec()



"""This module provides the functionalities needed for CAMELS to produce a
running python script for its protocols.

The overview of the protocol-file looks like this:\n
- import sys
- add path_to_camels, path_to_camels/nomad_camels, path to instruments to path
- `standard_string`
- `variable_string`
- `device_import_string`
- plot-string
- outer protocol string (things like plots from single steps)
- plan-string (including the inner plan, w/o start/stop and the outer)
- add-main-string (protocol or step specific things that should be done in the main function)
- `standard_run_string`
- further metadata (user / sample, file)
- `standard_start_string`
- set up databroker and progress bar
- `devices_string`
- `standard_start_string2`
- `save_string`
- `final_string`
- `standard_start_string3`

giving shortly:\n
- imports/variables...
- def protocol_plan_inner
- def protocol_plan
- def create_plots
- def steps_add_main (plots of steps, etc.)
- def uid_collector
- def run_protocol_main
- def main
- if __name__ == "__main__"

With the call chain as follows:\n
- if --> main
- main --> create_plots, steps_add_main, run_protocol_main
- run_protocol_main --> uid_collector, protocol_plan
- protocol_plan --> protocol_plan_inner
"""

import os.path
import copy

import pathlib

from nomad_camels.utility import variables_handling, load_save_functions

from nomad_camels.bluesky_handling.builder_helper_functions import plot_creator
from nomad_camels.utility import device_handling


# The default string in the beginning of the protocol including imports etc.
standard_string = "import numpy as np\n"
standard_string += "import importlib\n"
standard_string += "import bluesky\n"
standard_string += "import ophyd\n"
standard_string += "from bluesky import RunEngine\n"
standard_string += "from bluesky.callbacks.best_effort import BestEffortCallback\n"
standard_string += "import bluesky.plan_stubs as bps\n"
standard_string += "import databroker\n"
standard_string += "from PySide6.QtWidgets import QApplication, QMessageBox\n"
standard_string += "from PySide6.QtCore import QCoreApplication, QThread\n"
standard_string += "import datetime\n"
standard_string += "import subprocess\n"
standard_string += "from nomad_camels.main_classes import plot_pyqtgraph, list_plot\n"
standard_string += "from nomad_camels.utility import theme_changing\n"
standard_string += (
    "from nomad_camels.bluesky_handling.evaluation_helper import Evaluator\n"
)
standard_string += (
    "from nomad_camels.bluesky_handling import helper_functions, variable_reading\n"
)
standard_string += "from event_model import RunRouter\n"
standard_string += "darkmode = False\n"
standard_string += 'theme = "default"\n'
standard_string += 'protocol_step_information = {"protocol_step_counter": 0, "total_protocol_steps": 1, "protocol_stepper_signal": None}\n'


# this string is used for collection of uids to later access the data
standard_run_string = "uids = []\n"
standard_run_string += "def uid_collector(name, doc):\n"
standard_run_string += '\tuids.append(doc["uid"])\n\n\n'
standard_run_string += 'def run_protocol_main(RE, dark=False, used_theme="default", catalog=None, devices=None, md=None):\n'
standard_run_string += "\tdevs = devices or {}\n"
standard_run_string += "\tmd = md or {}\n"
standard_run_string += "\tglobal darkmode, theme, protocol_step_information\n"
standard_run_string += "\tdarkmode, theme = dark, used_theme\n"

# this is the string with the main function of the protocol, starting everything
# and also the if branch, whether it is the main script to execute everything
# without importing
standard_start_string = "\n\n\ndef main():\n"
standard_start_string += "\tRE = RunEngine()\n"
standard_start_string += "\tbec = BestEffortCallback()\n"
standard_start_string += "\tRE.subscribe(bec)\n"
standard_start_string2 = "\t\tplot_etc = create_plots(RE)\n"
standard_start_string2 += "\t\tadditional_step_data = steps_add_main(RE, devs)\n"
standard_start_string2 += "\t\tcreate_live_windows()\n"
standard_start_string2 += (
    "\t\trun_protocol_main(RE=RE, catalog=catalog, devices=devs, md=md)\n"
)
standard_start_string3 = '\n\n\nif __name__ == "__main__":\n'
standard_start_string3 += "\tmain()\n"
# standard_start_string3 += '\tapp = QCoreApplication.instance()\n'
standard_start_string3 += '\tprint("protocol finished!")\n'
standard_start_string3 += "\tif app is not None:\n"
standard_start_string3 += "\t\tsys.exit(app.exec())\n"
# standard_start_string += '\treturn plot_dat, additional_step_data\n'
standard_final_string = "\tfinally:\n"
standard_final_string += '\t\twhile RE.state not in ["idle", "panicked"]:\n'
standard_final_string += "\t\t\timport time\n"
standard_final_string += "\t\t\ttime.sleep(0.5)\n"

standard_nexus_dict = {
    "/ENTRY[entry]/operator/address": "metadata_start/user/Address (affiliation)",
    "/ENTRY[entry]/operator/affiliation": "metadata_start/user/Affiliation",
    "/ENTRY[entry]/operator/email": "metadata_start/user/E-Mail",
    "/ENTRY[entry]/operator/name": "metadata_start/user/Name",
    "/ENTRY[entry]/operator/orcid": "metadata_start/user/ORCID",
    "/ENTRY[entry]/operator/telephone_number": "metadata_start/user/Phone",
    "/ENTRY[entry]/start_time": "metadata_start/time",
    "/ENTRY[entry]/SAMPLE[sample]/data_identifier": "metadata_start/sample/Identifier",
    "/ENTRY[entry]/SAMPLE[sample]/sample_name": "metadata_start/sample/Name",
    "/ENTRY[entry]/SAMPLE[sample]/sample_history": "metadata_start/sample/Preparation-Info",
    "/ENTRY[entry]/PROCESS[process]/program": "metadata_start/program",
    "/ENTRY[entry]/PROCESS[process]/version": "metadata_start/version",
    "/ENTRY[entry]/SAMPLE[sample]/measured_data": "data",
}


def build_from_path(
    path, save_path="test.nxs", catalog="CAMELS_CATALOG", userdata=None, sampledata=None
):
    """Creating the runable python file from a given `protocol`.

    Parameters
    ----------
    path : str, path
        The path to the protocol that should be built.
    save_path : str, path
         (Default value = 'test.nxs')
         The path, where the data should be saved to.
    catalog : str
         (Default value = 'CAMELS_CATALOG')
         The name of the databroker catalog that should be used.
    userdata : dict, None
         (Default value = None)
         Metadata that describes the user.
    sampledata : dict, None
         (Default value = None)
         Metadata that describes the sample.
    """
    protocol = load_save_functions.load_protocol(path)
    path = pathlib.Path(path)
    build_protocol(
        protocol, path.with_suffix(".py"), save_path, catalog, userdata, sampledata
    )


def build_protocol(
    protocol,
    file_path,
    save_path="test.nxs",
    catalog="CAMELS_CATALOG",
    userdata=None,
    sampledata=None,
):
    """Creating the runable python file from a given `protocol`.

    Parameters
    ----------
    protocol : main_classes.protocol_class.Measurement_Protocol
        The protocol that provides the information for the python file.
    file_path : str, path
        The path, where the file should be saved.
    save_path : str, path
         (Default value = 'test.nxs')
         The path, where the data should be saved to.
    catalog : str
         (Default value = 'CAMELS_CATALOG')
         The name of the databroker catalog that should be used.
    userdata : dict, None
         (Default value = None)
         Metadata that describes the user.
    sampledata : dict, None
         (Default value = None)
         Metadata that describes the sample.
    """
    # the protocol is converted to a dictionary and saved as a json
    protocol_dict = load_save_functions.get_save_str(protocol)
    if not isinstance(file_path, pathlib.Path):
        file_path = pathlib.Path(file_path)
    cprot_path = file_path.with_suffix(".cprot").as_posix()
    load_save_functions.save_dictionary(cprot_path, protocol_dict)

    # first the path is prepared
    if not protocol.loop_steps:
        raise Exception(
            f'The protocol "{protocol.name}" is empty.\nYou cannot run or build an empty protocol.'
        )
    if not isinstance(save_path, pathlib.Path):
        save_path = pathlib.Path(save_path)
    if isinstance(save_path, pathlib.WindowsPath):
        save_path = save_path.as_posix()

    # clearing leftovers from former builds
    read_channel_names_old = list(variables_handling.read_channel_names)
    read_channel_sets_old = list(variables_handling.read_channel_sets)
    variables_handling.read_channel_names.clear()
    variables_handling.read_channel_sets.clear()

    # beginning of larger strings
    device_import_string = "\n"
    devices_string = ""
    read_device_drivers_string = ""
    variable_string = "\nnamespace = {}\n"
    variable_string += "all_fits = {}\n"
    variable_string += "plots = []\n"
    # variable_string += "plot_data = []\n"
    variable_string += "boxes = {}\n"
    variable_string += "live_windows = []\n"
    variable_string += "app = None\n"
    variable_string += f'save_path = "{save_path}"\n'
    variable_string += f'session_name = "{protocol.session_name}"\n'
    variable_string += f"export_to_csv = {protocol.export_csv}\n"
    variable_string += f"export_to_json = {protocol.export_json}\n"
    if "new_file_each_run" in variables_handling.preferences:
        variable_string += f'new_file_each_run = {variables_handling.preferences["new_file_each_run"]}\n'
    else:
        variable_string += f"new_file_each_run = False\n"
    variable_string += f"do_nexus_output = {protocol.use_nexus}\n"
    additional_string_devices = ""
    final_string = "\t\tfor name, device in devs.items():\n"
    final_string += '\t\t\tif hasattr(device, "finalize_steps") and callable(device.finalize_steps):\n'
    final_string += "\t\t\t\tdevice.finalize_steps()\n"

    # now all the variables of the protocol are added to the namespace
    # this includes also the variables of steps such as a foor-loop
    for var, val in protocol.variables.items():
        if variables_handling.check_data_type(val) == "String":
            val = f'"{val}"'
        if "(" in var or ")" in var:
            continue
        variable_string += f"{var} = {val}\n"
        variable_string += f'namespace["{var}"] = {var}\n'
    for var, val in protocol.loop_step_variables.items():
        if variables_handling.check_data_type(val) == "String":
            val = f'"{val}"'
        if "(" in var or ")" in var:
            continue
        variable_string += f"{var} = {val}\n"
        variable_string += f'namespace["{var}"] = {var}\n'
    variable_string += f'\n{protocol.name}_variable_signal = variable_reading.Variable_Signal(name="{protocol.name}_variable_signal", variables_dict=namespace)\n'
    # this handles all the used devices
    for dev in protocol.get_used_devices():
        device = variables_handling.devices[dev]
        classname = device.ophyd_class_name
        if protocol.skip_config:
            config = {}
        else:
            config = copy.deepcopy(device.get_config())
        settings = copy.deepcopy(device.get_settings())
        additional_info = copy.deepcopy(device.get_additional_info())
        # information on the connection is reshaped to be useful
        if "connection" in settings:
            conn = settings.pop("connection")
            if "type" in conn:
                conn.pop("type")
            settings.update(conn)
        if "idn" in settings:
            settings.pop("idn")

        # the non_string settings are strings in the dictionary, but are a known
        # variable or instance, so they are added without quotation marks
        extra_settings = {}
        non_strings = []
        for key in settings:
            if key.startswith("!non_string!_"):
                extra_settings[key.replace("!non_string!_", "")] = settings[key]
                non_strings.append(key)
        for s in non_strings:
            settings.pop(s)

        # same applies to non_string configs
        extra_config = {}
        non_strings = []
        for key in config:
            if key.startswith("!non_string!_"):
                extra_config[key.replace("!non_string!_", "")] = config[key]
                non_strings.append(key)
        for s in non_strings:
            config.pop(s)

        # Information about the device is written into the protocol as well
        additional_info["device_class_name"] = classname
        if "description" in additional_info:
            desc = additional_info["description"].replace("\n", "\n\t\t")
            devices_string += f'\t\t"""{dev} ({classname}):\n\t\t{desc}"""\n'
        devices_string += f"\t\tsettings = {settings}\n"
        devices_string += f"\t\tadditional_info = {additional_info}\n"
        devices_string += f'\t\t{dev} = {classname}("{dev}:", name="{dev}", '
        for key, value in extra_settings.items():
            devices_string += f"{key}={value}, "
        devices_string += "**settings)\n"

        # the devices are being connected, then their config is called and
        # written to the metadata
        devices_string += f'\t\tprint("connecting {dev}")\n'
        devices_string += f"\t\t{dev}.wait_for_connection()\n"
        devices_string += f"\t\tconfig = {config}\n"
        for k, v in extra_config.items():
            devices_string += f'\t\tconfig["{k}"] = {v}\n'
        devices_string += f"\t\tconfigs = {dev}.configure(config)[1]\n"
        devices_string += f'\t\tdevice_config["{dev}"] = {{"settings": {{}}}}\n'
        devices_string += f'\t\tdevice_config["{dev}"]["settings"].update(helper_functions.simplify_configs_dict(configs))\n'
        devices_string += f'\t\tdevice_config["{dev}"]["settings"].update(settings)\n'
        devices_string += f'\t\tdevice_config["{dev}"].update(additional_info)\n'
        devices_string += f'\t\tdevs.update({{"{dev}": {dev}}})\n'
        if device.name in device_handling.local_packages:
            device_import_string += f'sys.path.append(r"{device_handling.local_package_paths[device.name]}")\n'
        device_import_string += f"from nomad_camels_driver_{device.name}.{device.name}_ophyd import {classname}\n"
        additional_string_devices += device.get_additional_string()
        # For each device execute the get_opyd_and_py_file_contents from the helper_functions
        # This adds the .pya and .py files to the metadata dictionary md
        read_device_drivers_string += f"\tmd = helper_functions.get_opyd_and_py_file_contents({classname}, md, '{dev}')\n"

    # finishing up the device initialization
    devices_string += '\t\tprint("devices connected")\n'
    devices_string += f'\t\tmd = {{"devices": device_config}}\n'

    # the plots are created
    plot_string, plotting = plot_creator(protocol.plots, multi_stream=True)

    # everything is put together for the complete protocol-string
    protocol_string: str = "import sys\n"
    protocol_string += (
        f'sys.path.append(r"{os.path.dirname(variables_handling.CAMELS_path)}")\n'
    )
    protocol_string += f'sys.path.append(r"{os.path.dirname(variables_handling.CAMELS_path)}/nomad_camels")\n'
    protocol_string += (
        f'sys.path.append(r"{variables_handling.device_driver_path}")\n\n'
    )
    protocol_string += standard_string
    protocol_string += f"{variable_string}\n\n"
    protocol_string += device_import_string
    protocol_string += protocol.get_outer_string()
    protocol_string += protocol.get_plan_string()
    protocol_string += plot_string
    protocol_string += protocol.get_add_main_string()
    protocol_string += protocol.get_live_interaction_string()
    protocol_string += standard_run_string

    # setting up the progress bar
    protocol_string += f'\tprotocol_step_information["total_protocol_steps"] = {protocol.get_total_steps()}\n'
    protocol_string += additional_string_devices

    # add user / sample to metadata
    sampledata = sampledata or {"name": "default_sample"}
    userdata = userdata or {"name": "default_user"}
    protocol_string += user_sample_string(userdata, sampledata)
    protocol_string += f'\tmd["session_name"] = session_name\n'
    protocol_string += f'\tmd["protocol_overview"] = """{protocol.get_short_string().encode("unicode_escape").decode()}"""\n'
    protocol_string += f'\tmd["description"] = {repr(protocol.description)}\n'
    protocol_string += f'\tmd["measurement_tags"] = {protocol.tags}\n'
    protocol_string += (
        f'\tmd["measurement_description"] = {repr(protocol.measurement_description)}\n'
    )

    protocol_string += f"\ttry:\n"
    protocol_string += f'\t\twith open("{cprot_path}", "r", encoding="utf-8") as f:\n'
    protocol_string += '\t\t\tmd["protocol_json"] = f.read()\n'
    protocol_string += "\texcept FileNotFoundError:\n"
    protocol_string += "\t\tprint('Could not find protocol configuration file, information will be missing in data.')\n"

    # reading the file itself and adding it to the metadata
    protocol_string += '\twith open(__file__, "r", encoding="utf-8") as f:\n'
    protocol_string += '\t\tmd["python_script"] = f.read()\n'
    protocol_string += read_device_drivers_string
    protocol_string += '\tmd["variables"] = namespace\n'
    # protocol_string += '\tmd["plot_data"] = plot_data\n'

    # adding uid to RunEngine, calling the plan
    protocol_string += '\tsubscription_uid = RE.subscribe(uid_collector, "start")\n'
    protocol_string += f"\ttry:\n"
    protocol_string += f"\t\tRE({protocol.name}_plan(devs, md=md, runEngine=RE))\n"
    protocol_string += "\tfinally:\n"
    protocol_string += "\t\tRE.unsubscribe(subscription_uid)\n"
    protocol_string += "\t\tfor window in live_windows:\n"
    protocol_string += "\t\t\twindow.close()\n"

    # wait for RunEngine to finish, then save the data
    save_string = "\t\tif uids:\n"
    save_string += "\t\t\truns = catalog[tuple(uids)]\n"
    save_string += f"\t\t\thelper_functions.export_function(runs, save_path, {not protocol.h5_during_run}, new_file_each=new_file_each_run, plot_data=plots, do_nexus_output=do_nexus_output)\n"
    # save_string += (
    #     "\t\t\tbroker_to_NX(runs, save_path, plots,"
    #     "session_name=session_name,"
    #     "export_to_csv=export_to_csv,"
    #     "export_to_json=export_to_json,"
    #     "new_file_each_run=new_file_each_run)\n\n"
    # )
    if protocol.h5_during_run:
        save_string += "\t\tRE.unsubscribe(subscription_rr)\n"

    protocol_string += "\n\ndef ending_steps(runEngine, devs):\n"
    if protocol.use_end_protocol:
        protocol_string += sub_protocol_string(
            protocol_path=protocol.end_protocol, n_tabs=1
        )
    else:
        protocol_string += "\tyield from bps.checkpoint()\n"

    protocol_string += standard_start_string

    # checking for databroker catalog, using temp if not available
    protocol_string += "\ttry:\n"
    protocol_string += f'\t\tcatalog = databroker.catalog["{catalog}"]\n'
    protocol_string += "\texcept KeyError:\n"
    protocol_string += "\t\timport warnings\n"
    protocol_string += '\t\twarnings.warn("Could not find databroker catalog, using temporary catalog. If data is not transferred, it might get lost.")\n'
    protocol_string += "\t\tcatalog = databroker.temp().v2\n"
    protocol_string += "\tRE.subscribe(catalog.v1.insert)\n\n"

    # progress bar setup
    protocol_string += "\tfrom nomad_camels.utility import tqdm_progress_bar\n"
    protocol_string += (
        f"\ttqdm_bar = tqdm_progress_bar.ProgressBar({protocol.get_total_steps()})\n\n"
    )
    protocol_string += (
        '\tprotocol_step_information["protocol_stepper_signal"] = tqdm_bar\n'
    )

    # all the devices and the actual run are in a try-block
    protocol_string += "\tdevs = {}\n\tdevice_config = {}\n\ttry:\n"
    protocol_string += devices_string
    if protocol.h5_during_run:
        protocol_string += "\t\trr = RunRouter([lambda x, y: helper_functions.saving_function(x, y, save_path, new_file_each_run, plots, do_nexus_output)])\n"
        protocol_string += "\t\tsubscription_rr = RE.subscribe(rr)\n"
    protocol_string += standard_start_string2
    protocol_string += standard_final_string
    protocol_string += final_string
    protocol_string += save_string
    protocol_string += standard_start_string3
    protocol_string = protocol_string.expandtabs(4)

    # the string is written to the file
    if not os.path.isdir(os.path.dirname(file_path)):
        os.makedirs(os.path.dirname(file_path))
    with open(file_path, "w", encoding="utf-8") as file:
        file.write(protocol_string)

    variables_handling.read_channel_names = read_channel_names_old
    variables_handling.read_channel_sets = read_channel_sets_old


def user_sample_string(userdata, sampledata):
    """Returns the string adding userdata and sampledata to the md.

    Parameters
    ----------
    userdata : dict
        data on the user
    sampledata : dict
        data on the sample
    Returns
    -------
    u_s_string : str
        the string containing the data and adding it to the metadata
    """
    u_s_string = f'\tmd["user"] = {userdata}\n'
    u_s_string += f'\tmd["sample"] = {sampledata}\n'
    return u_s_string


def sub_protocol_string(
    protocol_path,
    n_tabs=1,
    variables_in=None,
    variables_out=None,
    data_output="sub-stream",
):
    """Overwrites the signal for the progressbar and the number of steps in
    the subprotocol's module. Evaluates the input variables, then writes
    them into the subprotocol's namespace and starts the subprotocol's
    _plan_inner function. Afterwards the output variables are written to the
    main namespace."""
    tabs = "\t" * n_tabs
    build_from_path(protocol_path)
    prot_name = os.path.basename(protocol_path)[:-6]
    protocol_string = (
        f"{tabs}{prot_name}_mod.protocol_step_information = protocol_step_information\n"
    )
    variables_in = variables_in or {"Variable": [], "Value": []}
    variables_out = variables_out or {"Variable": [], "Write to name": []}
    for i, var in enumerate(variables_in["Variable"]):
        protocol_string += (
            f'{tabs}{prot_name}_mod.{var} = eva.eval("{variables_in["Value"][i]}")\n'
        )
        protocol_string += f'{tabs}{prot_name}_mod.namespace["{var}"] = eva.eval("{variables_in["Value"][i]}")\n'
    protocol_string += (
        f"{tabs}{prot_name}_eva = Evaluator(namespace={prot_name}_mod.namespace)\n"
    )
    protocol_string += (
        f"{tabs}sub_eva_{prot_name} = runEngine.subscribe({prot_name}_eva)\n"
    )
    stream = prot_name
    if data_output == "main stream":
        stream = "primary"
    protocol_string += f'{tabs}yield from {prot_name}_mod.{prot_name}_plan_inner(devs, {prot_name}_eva, "{stream}", runEngine)\n'
    for i, var in enumerate(variables_out["Variable"]):
        protocol_string += f'{tabs}namespace["{variables_out["Write to name"][i]}"] = {prot_name}_mod.namespace["{var}"]\n'
    protocol_string += f"{tabs}runEngine.unsubscribe(sub_eva_{prot_name})\n"
    return protocol_string


def import_protocol_string(protocol_path, n_tabs=0):
    """Imports the subprotocol as <protocol_name>_mod."""
    tabs = "\t" * n_tabs
    prot_name = os.path.basename(protocol_path)[:-6]
    py_file = f"{protocol_path[:-6]}.py"
    if not os.path.isfile(py_file):
        sub_protocol = load_save_functions.load_protocol(protocol_path)
        build_protocol(sub_protocol, py_file)
    outer_string = f'{tabs}spec = importlib.util.spec_from_file_location("{prot_name}", "{py_file}")\n'
    outer_string += f"{tabs}{prot_name}_mod = importlib.util.module_from_spec(spec)\n"
    outer_string += f"{tabs}sys.modules[spec.name] = {prot_name}_mod\n"
    outer_string += f"{tabs}spec.loader.exec_module({prot_name}_mod)\n"
    return outer_string


def make_plots_string_of_protocol(
    protocol_path, use_own_plots=True, data_output="sub-stream", n_tabs=1
):
    """Creates the string to add the plots of the protocol to the main protocol.

    Parameters
    ----------
    protocol_path : str
        The path to the protocol that should be built.
    Returns
    -------
    plot_string : str
        The string that adds the plots of the protocol to the main protocol.
    """
    from nomad_camels.bluesky_handling import builder_helper_functions

    tabs = "\t" * n_tabs
    prot_name = os.path.basename(protocol_path)[:-6]
    plot_string = ""
    if use_own_plots:
        stream = f'"{prot_name}"'
        if data_output == "main stream":
            stream = '"primary"'
        plot_string += builder_helper_functions.get_plot_add_string(
            prot_name, stream, True, n_tabs
        )
    plot_string += f'{tabs}returner["{prot_name}_steps"] = {prot_name}_mod.steps_add_main(RE, devs)\n'
    return plot_string



"""A module used to update CAMELS. The import of nomad_camels is necessary,
since otherwise it would not show in the imported distributions for reading its
current version."""

import os
import sys
import subprocess
import re
import importlib
import nomad_camels  # has to be imported for the distribution version number!
from nomad_camels.ui_widgets import warn_popup

from PySide6.QtWidgets import (
    QMessageBox,
    QDialog,
    QGridLayout,
    QTextEdit,
    QDialogButtonBox,
    QPushButton,
)


def get_version():
    """checks the installed version of nomad-camels and returns it."""
    try:
        import pkg_resources

        return pkg_resources.get_distribution("nomad-camels").version
    except (AttributeError, pkg_resources.DistributionNotFound):
        try:
            return nomad_camels.__version__
        except Exception:
            return None


def get_latest_version():
    """Checks the latest version of nomad-camels and returns it."""
    import requests

    try:
        response = requests.get("https://pypi.org/pypi/nomad-camels/json", timeout=5)
        response.raise_for_status()
        data = response.json()
        return data["info"]["version"]
    except requests.exceptions.RequestException:
        completed_process = subprocess.run(
            [sys.executable, "-m", "pip", "install", "nomad-camels==random"],
            capture_output=True,
            text=True,
        )
        stderr_output = completed_process.stderr
        versions_string = re.search(r"\(from versions: (.*)\)", stderr_output)
        if versions_string:
            versions = versions_string.group(1).replace(" ", "").split(",")
            latest_version = versions[-1]
            return latest_version
        else:
            return None


def update_camels():
    """Calls a subprocess that updates CAMELS via pip."""
    from pathlib import Path
    from nomad_camels.utility.load_save_functions import appdata_path

    if sys.platform == "win32":
        exe = Path(sys.executable).resolve()
        if not str(exe)[-5] == "w":
            exew = Path(f"{str(exe)[:-4]}w.exe")
        else:
            exew = exe
            exe = Path(f"{str(exew)[:-5]}.exe")
        s = f"{exe} -m pip install --no-cache-dir nomad-camels --upgrade\n"
        s += f"{exew} "
        for a in sys.argv:
            s += f"{a} "
        s += "\nexit"
        f_path = Path(os.path.join(appdata_path, "camels_update.bat"))
        with open(f_path, "w+") as f:
            f.write(s)
        subprocess.Popen(["start", "cmd", "/k", f_path], shell=True)
        sys.exit()
    elif sys.platform.startswith("linux") or sys.platform == "darwin":
        exe = Path(sys.executable).resolve()
        s = f"{exe} -m pip install --no-cache-dir nomad-camels --upgrade\n"
        s += f"{exe} "
        for a in sys.argv:
            s += f"'{a}' "
        f_path = Path(os.path.join(appdata_path, "camels_update.sh"))
        with open(f_path, "w+") as f:
            f.write("#!/bin/bash\n")
            f.write(s)
        subprocess.check_call(["chmod", "+x", str(f_path)])
        subprocess.Popen(["x-terminal-emulator", "-e", "bash", "-c", f_path])
        sys.exit()


def question_message_box(parent=None):
    """
    Retrieves the currently installed version and the available version of
    nomad-camels. If the are the same, a message is displayed, and function
    exits. Otherwise, a question-box appears, whether the user wants to install
    the newer version of CAMELS. If yes, first `update_camels` and then
    `restart_camels` are called.

    Parameters
    ----------
    parent : QWidget
        (Default value = None)
        The parent widget to be used for the message boxes.
    """
    installed_version = get_version()
    available_version = get_latest_version()
    if installed_version == available_version:
        warn_popup.WarnPopup(
            parent,
            "Your version of NOMAD CAMELS is already up to date",
            "Already up to date",
            info_icon=True,
        )
        return
    update_dialog = QMessageBox.question(
        parent,
        "Update NOMAD CAMELS",
        f"Your current version of NOMAD CAMELS is {installed_version}\n"
        f"The newest available version is {available_version}\n"
        f"Do you want to update now? (NOMAD CAMELS will have to restart)",
        QMessageBox.Yes | QMessageBox.No,
    )
    if update_dialog != QMessageBox.Yes:
        return
    update_camels()
    restart_camels(parent)


def restart_camels(parent=None, ask_restart=True):
    """Restarts CAMELS. If `ask_restart`, a question-messagebox appears first,
    to make sure the user actually wants to restart.

    Parameters
    ----------
    parent : QWidget
        (Default value = None)
        The parent widget to be used for the message boxes.
    ask_restart : bool
        (Default value = True)
        If True, the user is asked whether to restart CAMELS.
    """
    if ask_restart:
        restart_dialog = QMessageBox.question(
            parent,
            "Restart NOMAD CAMELS now?",
            "Do you want to restart NOMAD CAMELS now?",
            QMessageBox.Yes | QMessageBox.No,
        )
        if restart_dialog != QMessageBox.Yes:
            return
    os.execl(sys.executable, sys.executable, *sys.argv)


def check_up_to_date():
    """Gets the installed and available version of CAMELS and checks whether
    they are the same. Returns the outcome as a bool."""
    installed_version = get_version()
    available_version = get_latest_version()
    return installed_version == available_version


def auto_update(parent):
    """
    Called if auto-update is set in the preferences. If the currently installed
    version is not up to date, `question_message_box` is called.

    Parameters
    ----------
    parent : QWidget
        (Default value = None)
        The parent widget to be used for the message boxes.
    """
    if not check_up_to_date():
        question_message_box(parent)


def show_release_notes():
    class MarkdownDialog(QDialog):
        def __init__(self, markdown_text):
            super().__init__()
            self.setWindowTitle("Changelog - NOMAD CAMELS")

            layout = QGridLayout(self)

            # Create a QTextEdit to display the Markdown content
            text_edit = QTextEdit(self)
            text_edit.setMarkdown(markdown_text)
            text_edit.setReadOnly(True)

            # Add OK button
            button_box = QDialogButtonBox(QDialogButtonBox.Ok, self)
            button_box.accepted.connect(self.accept)

            layout.addWidget(text_edit, 0, 0)
            layout.addWidget(button_box, 1, 0)

            self.setLayout(layout)
            self.setMinimumWidth(500)

    # read package's readme file
    try:
        with importlib.resources.open_text("nomad-camels", "README.md") as f:
            readme = f.read()
    except (FileNotFoundError, ModuleNotFoundError):
        try:
            with open("README.md", "r") as f:
                readme = f.read()
        except FileNotFoundError:
            try:
                with open("../README.md", "r") as f:
                    readme = f.read()
            except FileNotFoundError:
                return
    changelog = readme.split("# Changelog\n")[1]
    while changelog.startswith("\n"):
        changelog = changelog[1:]
    # newest_log = changelog.split("\n#")[0]
    dialog = MarkdownDialog(changelog)
    dialog.exec_()



import pytest
from pytest_mock import mocker
from unittest.mock import MagicMock
from nomad_camels.main_classes.device_class import Device, Device_Config, Local_VISA
from PySide6.QtWidgets import QApplication, QMainWindow


def test_device_initialization():
    device = Device(name="TestDevice", virtual=True, tags=["test", "virtual"])

    assert device.name == "TestDevice"
    assert device.virtual is True
    assert device.tags == ["test", "virtual"]
    assert device.settings == {}
    assert device.config == {}
    assert device.passive_config == {}
    assert device.channels == {}
    assert device.config_channels == {}
    assert device.non_channel_functions == []
    assert device.main_thread_only is False
    assert device.controls == {}


def test_get_channels(mocker):
    mock_device = MagicMock()
    mock_device.configuration_attrs = []
    mocker.patch(
        "nomad_camels.main_classes.device_class.OphydDevice", return_value=mock_device
    )
    mock_get_outputs = mocker.patch(
        "nomad_camels.main_classes.device_class.get_outputs", return_value=[]
    )
    mock_get_channels = mocker.patch(
        "nomad_camels.main_classes.device_class.get_channels", return_value=([], [])
    )

    device = Device(name="TestDevice")
    channels = device.get_channels()

    assert mock_get_outputs.called
    assert mock_get_channels.called
    assert channels == {}


def test_get_additional_info():
    device = Device(additional_info={"info_key": "info_value"})
    assert device.get_additional_info() == {"info_key": "info_value"}


def test_get_finalize_steps():
    device = Device(name="TestDevice")
    steps_str = device.get_finalize_steps()

    expected_str = "\t\tif 'TestDevice' in devs and hasattr(devs['TestDevice'], 'finalize_steps') and callable(devs['TestDevice'].finalize_steps):\n"
    expected_str += "\t\t\tdevs['TestDevice'].finalize_steps()\n"

    assert steps_str == expected_str


from nomad_camels.main_classes.device_class import (
    check_output,
    get_channels,
    get_outputs,
)
from ophyd import Signal, SignalRO


def test_check_output():
    assert check_output(Signal) is True
    assert check_output(SignalRO) is False


def test_get_outputs(mocker):
    mock_device = MagicMock()
    mock_device.walk_components.return_value = [
        MagicMock(item=MagicMock(attr="signal1", cls=MagicMock()))
    ]
    mocker.patch(
        "nomad_camels.main_classes.device_class.check_output", return_value=True
    )

    outputs = get_outputs(mock_device)
    assert "signal1" in outputs


def test_get_channels(mocker):
    mock_device = MagicMock()
    mock_device.walk_components.return_value = [
        MagicMock(item=MagicMock(attr="channel1", cls=MagicMock()))
    ]
    mock_device.configuration_attrs = []

    channels, config_channels = get_channels(
        mock_device, include_metadata=False, include_config=True
    )

    assert "channel1" in channels


@pytest.fixture(scope="module")
def app():
    """Fixture for creating a QApplication instance"""
    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    return app


@pytest.fixture
def main_window(app):
    """Fixture for creating a QMainWindow instance"""
    window = QMainWindow()
    window.show()
    return window


@pytest.fixture
def device_config(main_window):
    """Fixture for creating a Device_Config instance"""
    config = Device_Config(
        parent=main_window, device_name="Test Device", data="Test Data"
    )
    main_window.setCentralWidget(config)
    return config


def test_initialization(device_config):
    """Test the initialization of the Device_Config class"""
    assert device_config.lineEdit_custom_name.text() == "Test Data"
    assert device_config.textEdit_desc.toPlainText() == ""
    assert device_config.comboBox_connection_type.count() == 0


def test_adjust_text_edit_size(device_config):
    """Test the adjust_text_edit_size method"""
    initial_height = device_config.textEdit_desc.height()
    device_config.textEdit_desc.setText("Some text\nwith multiple lines.")
    device_config.adjust_text_edit_size()
    assert device_config.textEdit_desc.height() > initial_height


def test_connection_type_changed(device_config):
    """Test the connection_type_changed method"""
    device_config.comboBox_connection_type.addItem("Local VISA")
    device_config.comboBox_connection_type.setCurrentText("Local VISA")
    device_config.connection_type_changed()
    assert isinstance(device_config.connector, Local_VISA)


def test_get_settings(device_config):
    """Test the get_settings method"""
    device_config.comboBox_connection_type.addItem("Local VISA")
    device_config.comboBox_connection_type.setCurrentText("Local VISA")
    device_config.connection_type_changed()
    settings = device_config.get_settings()
    assert settings["connection"]["type"] == "Local VISA"


def test_load_settings(device_config):
    """Test the load_settings method"""
    device_config.comboBox_connection_type.addItem("Local VISA")
    device_config.settings_dict = {"connection": {"type": "Local VISA"}}
    device_config.load_settings()
    assert device_config.comboBox_connection_type.currentText() == "Local VISA"


def test_get_config(device_config):
    """Test the get_config method"""
    assert device_config.get_config() == {}


def test_get_info(device_config):
    """Test the get_info method"""
    description = "This is a test device."
    device_config.textEdit_desc.setText(description)
    info = device_config.get_info()
    assert info["description"] == description



import os
import platform, subprocess

from PySide6.QtWidgets import (
    QApplication,
    QDialog,
    QLabel,
    QLineEdit,
    QGridLayout,
    QDialogButtonBox,
    QComboBox,
    QPushButton,
)
from PySide6.QtGui import QPixmap, Qt

from importlib import resources
from nomad_camels import graphics

from nomad_camels.utility import variables_handling

nomad_url = "https://nomad-lab.eu/prod/v1/staging/gui/analyze/apis"


class LoginDialog(QDialog):
    """UI widget to handle the login to NOMAD."""

    def __init__(self, parent=None):
        super().__init__(parent)

        info_label = QLabel(
            "We strongly recommend to use NOMAD's app token for connecting.\nCopy the token from NOMAD (button below)."
        )

        oasis_url = ""
        if "NOMAD_URL" in variables_handling.preferences:
            oasis_url = variables_handling.preferences["NOMAD_URL"]

        self.comboBox_nomad_choice = QComboBox()
        self.comboBox_nomad_choice.addItems(["central NOMAD", "NOMAD Oasis"])

        self.pushButton_token = QPushButton("get the\ntoken!")
        self.pushButton_token.clicked.connect(self.open_token_link)

        if oasis_url:
            self.comboBox_nomad_choice.setCurrentText("NOMAD Oasis")

        self.label_oasis_url = QLabel("Oasis URL:")
        self.lineEdit_oasis_url = QLineEdit(oasis_url)

        self.label_logo = QLabel()
        self.label_logo.setAlignment(Qt.AlignCenter)

        self.label_info_oasis = QLabel(
            "Hint: You can set the URL for the Oasis also in the settings of CAMELS.\nCurrently we only support direct login for NOMAD Oasis via username/password.\nIf you use Single-Sing-On for your Oasis, you need to use NOMAD's app token."
        )
        self.label_auth_type = QLabel("Authentication type:")
        self.comboBox_auth_type = QComboBox()
        self.comboBox_auth_type.addItems(["user/password", "token"])
        self.comboBox_auth_type.setCurrentText("token")
        self.comboBox_auth_type.currentTextChanged.connect(self.change_login)

        self.username_label = QLabel("Username/email:")
        self.username_input = QLineEdit()
        self.password_label = QLabel("Password:")
        self.password_input = QLineEdit()
        self.password_input.setEchoMode(QLineEdit.Password)

        self.label_token = QLabel("Authentication token:")
        self.lineEdit_token = QLineEdit()
        self.lineEdit_token.setEchoMode(QLineEdit.Password)

        self.username = None
        self.password = None

        self.button_box = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        layout = QGridLayout()
        layout.addWidget(self.comboBox_nomad_choice, 10, 0, 1, 2)
        layout.addWidget(self.label_oasis_url, 11, 0)
        layout.addWidget(self.lineEdit_oasis_url, 11, 1)
        layout.addWidget(self.label_logo, 12, 0)
        layout.addWidget(self.pushButton_token, 12, 1)

        layout.addWidget(self.label_info_oasis, 15, 0, 1, 2)

        layout.addWidget(info_label, 0, 0, 1, 2)
        layout.addWidget(self.label_auth_type, 1, 0)
        layout.addWidget(self.comboBox_auth_type, 1, 1)

        layout.addWidget(self.username_label, 5, 0)
        layout.addWidget(self.username_input, 5, 1)
        layout.addWidget(self.password_label, 6, 0)
        layout.addWidget(self.password_input, 6, 1)

        layout.addWidget(self.label_token, 7, 0)
        layout.addWidget(self.lineEdit_token, 7, 1)

        layout.addWidget(self.button_box, 20, 0, 1, 2)

        self.setLayout(layout)
        self.setWindowTitle("Log in to NOMAD")
        self.change_login()
        self.url = None
        self.token = None

        self.comboBox_nomad_choice.currentTextChanged.connect(self.change_login)

    def open_token_link(self):
        if self.comboBox_nomad_choice.currentText() == "NOMAD Oasis":
            path = self.lineEdit_oasis_url.text()
            if "/api/" in path:
                path = path.split("/api/")[0]
            elif path.endswith("/api"):
                path = path.split("/api")[0]
            if "/gui/" in path:
                path = path.split("/gui/")[0]
            elif path.endswith("/gui"):
                path = path.split("/gui")[0]
            if path.endswith("/"):
                path = path[:-1]
            path = f"{path}/gui/analyze/apis"
        else:
            path = nomad_url
        if platform.system() == "Windows":
            os.startfile(path)
        else:
            opener = "open" if platform.system() == "Darwin" else "xdg-open"
            subprocess.call([opener, path])

    def accept(self):
        oasis = self.comboBox_nomad_choice.currentText() == "NOMAD Oasis"
        use_token = self.comboBox_auth_type.currentText() == "token"
        self.username = self.username_input.text()
        self.password = self.password_input.text()
        if oasis:
            self.url = self.lineEdit_oasis_url.text()
        if use_token:
            self.token = self.lineEdit_token.text()
            if "Bearer%20" in self.token:
                self.token = self.token.split("Bearer%20")[-1]
            self.token = self.token.replace('"', "")
        super().accept()

    def change_login(self):
        oasis = self.comboBox_nomad_choice.currentText() == "NOMAD Oasis"
        image = QPixmap()
        if oasis:
            image.load(str(resources.files(graphics) / "oasis-horizontal.png"))
        else:
            image.load(str(resources.files(graphics) / "nomad-horizontal.png"))
        self.label_logo.setPixmap(image)

        self.label_oasis_url.setHidden(not oasis)
        self.lineEdit_oasis_url.setHidden(not oasis)
        # self.comboBox_auth_type.setHidden(not oasis)
        # self.label_auth_type.setHidden(not oasis)
        self.label_info_oasis.setHidden(not oasis)

        use_token = self.comboBox_auth_type.currentText() == "token"
        self.username_input.setHidden(use_token)
        self.username_label.setHidden(use_token)
        self.password_input.setHidden(use_token)
        self.password_label.setHidden(use_token)
        self.label_token.setHidden(not use_token)
        self.lineEdit_token.setHidden(not use_token)


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    dialog = LoginDialog()
    dialog.exec()



import importlib
from PySide6.QtWidgets import QComboBox, QLabel

from nomad_camels.main_classes.loop_step import Loop_Step, Loop_Step_Config
from nomad_camels.main_classes.device_class import Device_Config_Sub

from nomad_camels.utility import variables_handling


class Change_DeviceConf(Loop_Step):
    """
    With this step, one can change the config-attributes of a device.

    Attributes
    ----------
    device : str
        The name of the device that has the config changed.
    config_dict : dict
        The dictionary with the new config-attributes of the device.
    """

    def __init__(self, name="", parent_step=None, step_info=None, **kwargs):
        super().__init__(name, parent_step, step_info, **kwargs)
        self.step_type = "Change Device Config"
        if step_info is None:
            step_info = {}
        self.device = step_info["device"] if "device" in step_info else ""
        self.config_dict = (
            step_info["config_dict"] if "config_dict" in step_info else {}
        )
        self.settings_dict = (
            step_info["settings_dict"] if "settings_dict" in step_info else {}
        )

    def update_used_devices(self):
        """Includes self.device to the used devices."""
        if self.device == "advanced configuration":
            self.used_devices = []
            for conf in self.config_dict["channel"]:
                dev = variables_handling.config_channels[conf].device
                if dev not in self.used_devices:
                    self.used_devices.append(dev)
        else:
            self.used_devices = [self.device]

    def get_protocol_string(self, n_tabs=1):
        """Creates an instance of `self.device` and uses `self.config_dict` to
        get a complete config_dict from the instance.
        The string consists of calling `dev.configure(config)`, where config is
        the config_dict."""
        tabs = "\t" * n_tabs
        dev_name = self.device
        protocol_string = super().get_protocol_string(n_tabs)
        if dev_name != "advanced configuration":
            device = variables_handling.devices[dev_name]
            dev_type = device.name
            py_package = importlib.import_module(
                f"nomad_camels_driver_{dev_type}.{dev_type}"
            )
            dev_instance = py_package.subclass()
            dev_instance.config = self.config_dict
            config_dict = dev_instance.get_config()
            extra_config = {}
            non_strings = []
            for key in config_dict:
                if key.startswith("!non_string!_"):
                    extra_config[key.replace("!non_string!_", "")] = config_dict[key]
                    non_strings.append(key)
            for s in non_strings:
                config_dict.pop(s)
            config_dict.update(extra_config)
            protocol_string += f"{tabs}config = {config_dict}\n"
            protocol_string += f'{tabs}devs["{self.device}"].configure(config)\n'
        else:
            self.update_used_devices()
            for dev in self.used_devices:
                protocol_string += f"{tabs}config_dict = " + "{"
                for i, conf in enumerate(self.config_dict["channel"]):
                    dev_name, conf_name = variables_handling.config_channels[
                        conf
                    ].name.split(".")
                    if dev_name != dev:
                        continue
                    if conf not in variables_handling.config_channels:
                        raise Exception(
                            f"Trying to configure {conf} in {self.full_name}, but it does not exist!"
                        )
                    val = self.config_dict["value"][i].replace('"', '\\"')
                    protocol_string += f'"{conf_name}": eva.eval("{val}"), '
                protocol_string = protocol_string[:-2] + "}\n"
                protocol_string += f'{tabs}devs["{dev}"].configure(config_dict)\n'
        return protocol_string

    def get_protocol_short_string(self, n_tabs=0):
        """Includes the configured device in the short string."""
        short_string = super().get_protocol_short_string(n_tabs)
        short_string = f"{short_string[:-1]} - {self.device}\n"
        return short_string


class Change_DeviceConf_Config(Loop_Step_Config):
    """ """

    def __init__(self, loop_step: Change_DeviceConf, parent=None):
        super().__init__(parent, loop_step)
        self.loop_step = loop_step
        label_dev = QLabel("Device:")
        self.comboBox_device = QComboBox()
        devs = list(variables_handling.devices.keys()) + ["advanced configuration"]
        self.comboBox_device.addItems(devs)
        if loop_step.device in devs:
            self.comboBox_device.setCurrentText(loop_step.device)

        self.config_widget = Device_Config_Sub(parent=self)

        self.layout().addWidget(label_dev, 1, 0)
        self.layout().addWidget(self.comboBox_device, 1, 1)
        self.layout().addWidget(self.config_widget, 2, 0, 1, 5)
        self.comboBox_device.currentTextChanged.connect(self.device_changed)
        self.device_changed()

    def device_changed(self):
        """ """
        dev_name = self.comboBox_device.currentText()
        if not dev_name:
            return
        if dev_name == "advanced configuration":
            from nomad_camels.ui_widgets.channels_check_table import (
                Channels_Check_Table,
            )

            if self.loop_step.device == "advanced configuration":
                config_widge = Channels_Check_Table(
                    self,
                    ["set", "config", "value"],
                    True,
                    self.loop_step.config_dict,
                    [2],
                    use_configs=True,
                )
            else:
                config_widge = Channels_Check_Table(
                    self,
                    ["set", "config", "value"],
                    True,
                    {"channel": [], "value": []},
                    [2],
                    use_configs=True,
                )
        else:
            device = variables_handling.devices[dev_name]
            dev_type = device.name
            py_package = importlib.import_module(
                f"nomad_camels_driver_{dev_type}.{dev_type}"
            )
            if dev_name == self.loop_step.device:
                settings = self.loop_step.settings_dict
                config = self.loop_step.config_dict
            else:
                settings = dict(device.settings)
                config = dict(device.config)
            try:
                config_widge = py_package.subclass_config_sub(
                    parent=self, settings_dict=settings, config_dict=config
                )
            except AttributeError:
                try:
                    widge = py_package.subclass_config(
                        parent=self, settings_dict=settings, config_dict=config
                    )
                    config_widge = widge.sub_widget
                except AttributeError:
                    config_widge = Device_Config_Sub(parent=self)
        self.layout().replaceWidget(self.config_widget, config_widge)
        self.config_widget.deleteLater()
        self.config_widget = config_widge

    def update_step_config(self):
        """ """
        super().update_step_config()
        dev_name = self.comboBox_device.currentText()
        self.loop_step.device = dev_name
        if dev_name == "advanced configuration":
            self.loop_step.config_dict = self.config_widget.get_info()
        else:
            self.loop_step.config_dict = self.config_widget.get_config()
            self.loop_step.settings_dict = self.config_widget.get_settings()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'driver_builder.ui'
##
## Created by: Qt User Interface Compiler version 6.5.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QFrame,
    QGridLayout,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QWidget,
)


class Ui_VISA_Device_Builder(object):
    def setupUi(self, VISA_Device_Builder):
        if not VISA_Device_Builder.objectName():
            VISA_Device_Builder.setObjectName("VISA_Device_Builder")
        VISA_Device_Builder.resize(1141, 310)
        self.gridLayout = QGridLayout(VISA_Device_Builder)
        self.gridLayout.setObjectName("gridLayout")
        self.lineEdit_name = QLineEdit(VISA_Device_Builder)
        self.lineEdit_name.setObjectName("lineEdit_name")
        self.lineEdit_name.setMaximumSize(QSize(350, 16777215))

        self.gridLayout.addWidget(self.lineEdit_name, 0, 1, 1, 2)

        self.lineEdit_search_tags = QLineEdit(VISA_Device_Builder)
        self.lineEdit_search_tags.setObjectName("lineEdit_search_tags")
        self.lineEdit_search_tags.setMaximumSize(QSize(250, 16777215))

        self.gridLayout.addWidget(self.lineEdit_search_tags, 2, 1, 1, 1)

        self.label_baud_rate = QLabel(VISA_Device_Builder)
        self.label_baud_rate.setObjectName("label_baud_rate")

        self.gridLayout.addWidget(self.label_baud_rate, 3, 4, 1, 1)

        self.label = QLabel(VISA_Device_Builder)
        self.label.setObjectName("label")
        self.label.setMaximumSize(QSize(100, 16777215))

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.label_write_term = QLabel(VISA_Device_Builder)
        self.label_write_term.setObjectName("label_write_term")
        self.label_write_term.setMaximumSize(QSize(140, 16777215))

        self.gridLayout.addWidget(self.label_write_term, 2, 4, 1, 1)

        self.label_5 = QLabel(VISA_Device_Builder)
        self.label_5.setObjectName("label_5")
        self.label_5.setMaximumSize(QSize(100, 16777215))

        self.gridLayout.addWidget(self.label_5, 2, 0, 1, 1)

        self.label_read_term = QLabel(VISA_Device_Builder)
        self.label_read_term.setObjectName("label_read_term")
        self.label_read_term.setMaximumSize(QSize(140, 16777215))

        self.gridLayout.addWidget(self.label_read_term, 1, 4, 1, 1)

        self.lineEdit_write_term = QLineEdit(VISA_Device_Builder)
        self.lineEdit_write_term.setObjectName("lineEdit_write_term")

        self.gridLayout.addWidget(self.lineEdit_write_term, 2, 5, 1, 3)

        self.label_7 = QLabel(VISA_Device_Builder)
        self.label_7.setObjectName("label_7")
        self.label_7.setMaximumSize(QSize(100, 16777215))

        self.gridLayout.addWidget(self.label_7, 2, 2, 1, 1)

        self.lineEdit_baud_rate = QLineEdit(VISA_Device_Builder)
        self.lineEdit_baud_rate.setObjectName("lineEdit_baud_rate")

        self.gridLayout.addWidget(self.lineEdit_baud_rate, 3, 5, 1, 3)

        self.checkBox_VISA = QCheckBox(VISA_Device_Builder)
        self.checkBox_VISA.setObjectName("checkBox_VISA")

        self.gridLayout.addWidget(self.checkBox_VISA, 0, 4, 1, 4)

        self.lineEdit_read_term = QLineEdit(VISA_Device_Builder)
        self.lineEdit_read_term.setObjectName("lineEdit_read_term")

        self.gridLayout.addWidget(self.lineEdit_read_term, 1, 5, 1, 3)

        self.line = QFrame(VISA_Device_Builder)
        self.line.setObjectName("line")
        self.line.setFrameShape(QFrame.VLine)
        self.line.setFrameShadow(QFrame.Sunken)

        self.gridLayout.addWidget(self.line, 0, 3, 4, 1)

        self.retranslateUi(VISA_Device_Builder)

        QMetaObject.connectSlotsByName(VISA_Device_Builder)

    # setupUi

    def retranslateUi(self, VISA_Device_Builder):
        VISA_Device_Builder.setWindowTitle(
            QCoreApplication.translate("VISA_Device_Builder", "Form", None)
        )
        self.label_baud_rate.setText(
            QCoreApplication.translate(
                "VISA_Device_Builder", "Default Baud-Rate:", None
            )
        )
        self.label.setText(
            QCoreApplication.translate("VISA_Device_Builder", "Name:", None)
        )
        self.label_write_term.setText(
            QCoreApplication.translate(
                "VISA_Device_Builder", "Default Write-Termination:", None
            )
        )
        self.label_5.setText(
            QCoreApplication.translate("VISA_Device_Builder", "Search-Tags:", None)
        )
        self.label_read_term.setText(
            QCoreApplication.translate(
                "VISA_Device_Builder", "Default Read-Termination:", None
            )
        )
        self.lineEdit_write_term.setText(
            QCoreApplication.translate("VISA_Device_Builder", "\\r\\n", None)
        )
        self.label_7.setText(
            QCoreApplication.translate("VISA_Device_Builder", "(space-separated)", None)
        )
        self.lineEdit_baud_rate.setText(
            QCoreApplication.translate("VISA_Device_Builder", "9600", None)
        )
        self.checkBox_VISA.setText(
            QCoreApplication.translate("VISA_Device_Builder", "VISA connection", None)
        )
        self.lineEdit_read_term.setText(
            QCoreApplication.translate("VISA_Device_Builder", "\\r\\n", None)
        )

    # retranslateUi



from PySide6.QtCore import Signal, Qt
from PySide6.QtGui import QPainter
from PySide6.QtWidgets import (
    QWidget,
    QFrame,
    QLabel,
    QFrame,
    QPushButton,
    QCheckBox,
    QHBoxLayout,
    QListWidget,
    QListWidgetItem,
    QAbstractItemView,
    QMessageBox,
)


class RunWidget(QWidget):
    """
    A widget that represents a single run in the run queue. It contains a label with the name of the protocol, a checkbox whether the protocol is ready to run, and a button to remove the protocol from the queue.

    Parameters
    ----------
    text : str
        The name of the protocol / the text to display on the label."""

    def __init__(self, text):
        super().__init__()
        self.name = text

        self.layout = QHBoxLayout(self)

        self.checkbox = QCheckBox("ready")
        self.checkbox.setToolTip(
            "The next run will start automatically if it is set as ready.\nOtherwise the execution will pause."
        )
        self.label = QLabel(text)
        self.button = QPushButton("remove")
        self.button.setMaximumWidth(80)

        self.layout.addWidget(self.label)
        self.layout.addWidget(self.checkbox)
        self.layout.addWidget(self.button)


class RunQueue(QListWidget):
    """
    A QListWidget that represents the run queue. It contains a list of RunWidgets, each representing a single run in the queue. When a run is removed from the queue, the corresponding RunWidget is removed from the QListWidget. The order of the runs in the queue is determined by the order of the RunWidgets in the QListWidget. The RunQueue emits a signal when a the ready status of a run is changed, so that the next run can be started automatically.

    Each run can have its own variables, the values do not change the default settings for the protocol.
    """

    protocol_signal = Signal(str, dict, str)

    def __init__(self, parent=None, protocols_dict=None, variable_table=None):
        super().__init__(parent=parent)

        self.setDragDropMode(QAbstractItemView.InternalMove)

        self.order_list = []
        self.protocol_name_variables = {}

        # Connect the rowsMoved signal to the update_order_list slot
        self.model().rowsMoved.connect(self.update_order_list)
        self.protocols_dict = protocols_dict or {}
        self.variable_table = variable_table
        self.selectionModel().selectionChanged.connect(self.change_variable_table)
        self.last_selected = None

    def add_item(self, text, api_uuid=None):
        """
        Add a new item to the run queue. The item is represented by a RunWidget, which is added to a QListWidgetItem, which is then added to the QListWidget. The item is also added to the order list, which determines the order of the runs in the queue. The name of the protocol and its variables are stored in the protocol_name_variables dictionary.
        """
        item = QListWidgetItem(self)

        # Create a QFrame
        frame = QFrame()
        frame.setFrameShape(QFrame.StyledPanel)
        frame.setFrameShadow(QFrame.Sunken)

        # Create the run widget and add it to the frame
        item_widget = RunWidget(text)
        layout = QHBoxLayout(frame)
        layout.addWidget(item_widget)

        item.setSizeHint(frame.sizeHint())
        self.addItem(item)
        self.setItemWidget(item, frame)

        item_widget.button.clicked.connect(lambda: self.remove_item(item))
        item_widget.checkbox.clicked.connect(self.check_next_protocol)
        item_widget.button.clicked.connect(self.update_order_list)

        # Add the item to the order list
        self.order_list.append(item)
        self.protocol_name_variables[str(item)] = [
            text,
            self.protocols_dict[text].variables,
            api_uuid,
        ]

    def check_checkbox(self, name):
        """
        Check the checkbox of the specified item.

        Parameters
        ----------
        name : str
        string of the QListWidgetItem. The item to check.
        """
        for item in self.order_list:
            if name == str(item):
                widget = self.itemWidget(item)
                checkbox = widget.layout().itemAt(0).widget().checkbox
                checkbox.setChecked(True)
                checkbox.clicked.emit()

    def update_variables_queue(self, name, variables, index):
        """
        Update the variables for a protocol in the queue.

        Parameters
        ----------
        name : str
            The name of the protocol.
        variables : dict
            The new variables for the protocol.
        """
        item = self.order_list[index]
        if name == str(item):
            pass
        else:
            raise ValueError(
                "The name of the protocol does not match the name of the protocol at this index in the queue."
            )
        self.protocol_name_variables[str(item)][1] = variables
        self.change_variable_table()

    def check_next_protocol(self):
        """
        Check if the first protocol in the queue is ready to run. If it is, emit the protocol_signal with the name of the protocol and its variables. If the protocol is not ready to run, do nothing.

        Returns
        -------
        bool
            True if the next protocol is ready to run, False otherwise.
        """
        # Check if the next protocol is ready to run
        if len(self.order_list) == 0:
            return False

        # Get the first item in the order list
        item = self.order_list[0]
        widget = self.itemWidget(item)
        checkbox = widget.layout().itemAt(0).widget().checkbox

        # If the checkbox is checked, emit the run signal
        if checkbox.isChecked():
            self.change_variable_table()
            if self.protocol_name_variables[str(item)][2] is not None:
                from nomad_camels.api.api import write_protocol_result_path_to_db

                write_protocol_result_path_to_db(
                    self.protocol_name_variables[str(item)][2], "currently running"
                )
            self.protocol_signal.emit(*self.protocol_name_variables[str(item)])
            return True
        return False

    def remove_first(self):
        """
        Remove the first item in the queue.
        """
        if len(self.order_list) == 0:
            return
        self.remove_item(self.order_list[0], ask=False)

    def remove_item(self, item, ask=True):
        """
        Remove an item from the run queue. The item is removed from the order list and the QListWidget. If `ask` is True, a QMessageBox is shown to ask the user if they are sure they want to remove the item.

        Parameters
        ----------
        item : QListWidgetItem
            The item to remove.
        ask : bool, optional
            Whether to ask the user if they are sure they want to remove the item. The default is True.
        """
        # ask the user if they are sure
        name = self.itemWidget(item).layout().itemAt(0).widget().name
        if ask:
            dialog = QMessageBox()
            dialog.setWindowTitle("Remove run")
            dialog.setText(f"Remove {name} from the queue?")

            dialog.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
            dialog.setDefaultButton(QMessageBox.No)
            if dialog.exec() == QMessageBox.No:
                return

        # Remove the item from the order list
        self.protocol_name_variables.pop(str(item))
        self.order_list.remove(item)

        # Remove the item from the QListWidget
        self.takeItem(self.row(item))

    def remove_item_by_name(self, name):
        """
        Remove an item from the run queue by name.

        Parameters
        ----------
        name : str
            The name of the item to remove.
        """
        for item in self.order_list:
            if name == str(item):
                self.remove_item(item, ask=False)

    def update_order_list(self):
        """
        Update the order list to match the current order of items in the QListWidget

        """
        self.order_list = [self.item(i) for i in range(self.count())]
        self.check_next_protocol()

    def change_variable_table(self):
        """
        Change the variables displayed in the variable table to match the selected run.
        """
        if (
            self.last_selected is not None
            and str(self.last_selected) in self.protocol_name_variables
        ):
            self.protocol_name_variables[str(self.last_selected)][
                1
            ] = self.variable_table.update_variables()
        self.variable_table.editable_names = False
        self.variable_table.clear()
        # Get the current item
        self.last_selected = self.currentItem()
        if self.last_selected is None:
            self.variable_table.setHidden(True)
            return
        # Get the variables for the current item
        (
            name,
            variables,
        ) = self.protocol_name_variables[
            str(self.last_selected)
        ][:2]
        self.variable_table.protocol = self.protocols_dict[name]
        for var in variables:
            self.variable_table.append_variable(var, str(variables[var]), unique=False)
        self.variable_table.setHidden(False)
        self.variable_table.resizeColumnsToContents()

    def paintEvent(self, event):
        super().paintEvent(event)
        if self.count() == 0:
            painter = QPainter(self.viewport())
            painter.setPen(Qt.gray)

            # Set the font to be larger and bold
            font = painter.font()
            font.setPointSize(16)  # Adjust the size as needed
            font.setBold(True)
            painter.setFont(font)

            painter.drawText(self.rect(), Qt.AlignCenter, "Measurement Queue")
            painter.end()


if __name__ == "__main__":
    from PySide6.QtWidgets import QApplication

    app = QApplication([])

    widget = RunQueue()
    widget.show()

    app.exec()
